#!/usr/bin/perl -U
#============================================================================#
# vhost - one-step solution for all virtual hosting needs                    #
#         vhost.pl - core module, host/dns configuration module              #
#                                                                            #
# Copyright(c) Chaogic Systems, LLC. http://chaogic.com                      #
# Author: Jake Fan <jake@chaogic.com>                                        #
#                                                                            #
# This is Free Software;  permission to use, copy, modify, and distribute    #
# this software and its documentation for any purpose - with or without fee  #
# - is hereby granted, provided that the above copyright information and     #
# this permission notice appear in all copies and documentations.  This      #
# software is provided "as is" without express or implied warranty.  For     #
# more details, see GNU General Public License as published by the Free      #
# Software Foundation.                                                       #
#                                                                            #
#============================================================================#
#
$SUPPORT = <<'#';
# Linux, NetBSD, FreeBSD, OpenBSD
#
$REQUIRE = <<'#';
# /etc/fstab, /etc/shells, false, date, echo, cat, cd, pwd, dirname, basename, 
# uname -n, ls -l, id -G, du -sk, ln -sf, cp -pR, mv, rm -r, mkdir -p, 
# chown -R, chmod, touch, passwd, useradd -cdgsuo, usermod -cdgsuo, userdel, 
# groupadd -go, groupmod -go, groupdel, pw, gzip -f, gunzip -cf, test, 
# make, gcc, g++
#
$HISTORY = <<'#';
# Ver  Date    Credits         Comments
# ---- ------- --------------- -----------------------------------------------
# 0.90 09Nov00 Jake Fan        + auto ip aliasing and auto selection between 
#                                ip and name based hosting
# 1.00 27Nov00 Jake Fan        + virtual email account/alias management
# 1.10 10Dec00 Jake Fan        + virtual host alias management
# 1.20 24Jan01 Jake Fan        + secure setuid wrapper for unprivileged 
#                                administration and web integration
# 1.21 24Jan01 Jake Fan        + command history logging
# 1.30 06Feb01 Jake Fan        + modularized configuration files
#                              + support for arbitrary virtual services via 
#                                configuration modules
#                              + --version
# 1.31 09Feb01 Jake Fan        + distro-specific improvements
# 1.32 19Feb01 Jake Fan        + support for "user@domain.com" as pop3 login
# 1.33 25Feb01 Jake Fan        + proftpd configuration module
# 1.40 18Mar01 Jake Fan        + changed location for virtual host log files
#                              + --adddns
#                              + --deldns
# 1.41 26Mar01 Jake Fan        + improved --reload and --listhost output
# 1.42 05Apr01 Jake Fan        + --listhost caching for fast output
# 1.43 08Apr01 Jake Fan        + --uninstall
# 1.44 20Apr01 Jake Fan        + --cmdlog
#                              + --history
# 1.45 21Apr01 Jake Fan        + virtual root user
# 2.00 14May01 Jake Fan        + web gui
#                              + optimizations for web integration
# 2.01 07Jun01 Jake Fan        + support for dns sub-domains
#                              + catch-all email alias
# 3.00 30Oct02 Jake Fan        + completely redesigned and rewritten
#                              + fully modularized architecture
#                              + unlimited system profiles (configurations)
#                              + distributed system (clustering)
#                              + cross-platform support (linux/*bsd)
#                              + four levels of login (root/vroot/owner/vuser)
#                              + comprehensive quota system
#                              + built-in log rotation
#                              + virtual user home/web space
#                              + improved ip address management
#                              + improved web gui
#                              + --support
#                              + --require
#                              + postfix configuration module
#                              + imap (squirrelmail) configuration module
# 3.01 25Dec02 Jake Fan        * see change log for details
# 3.02 31Dec02 Jake Fan
# 3.03 03Jan03 Jake Fan
# 3.04 08Jan03 Jake Fan
# 3.05 31May03 Jake Fan
#
$VERSION = "vhost 3.05r2 Copyright(c) 2000-2003 Chaogic Systems, LLC. by Jake Fan\n";

$HELP = <<'#';
Add/Del vRoot: vhost --addvroot profile_dir [vroot_name] [-c] [-f]
               vhost --delvroot profile_dir [vroot_name] [-f]
  Add/Del DNS: vhost --adddns host_name [ip_address] [-z] [-i] [-q] [-f]
               vhost --deldns host_name [ip_address] [-q] [-f]
 Add/Del Host: vhost --addhost host_name owner_name [ip_address] [-q] [-f]
               vhost --delhost host_name [-q] [-f [-d]]
   Host Alias: vhost --addhali host_name host_alias [-q]
               vhost --delhali host_alias [-q]
 Add/Del User: vhost --adduser host_name user_name [full_name] [-f]
               vhost --deluser host_name user_name [-f [-d]]
   User Alias: vhost --adduali host_name user_alias|catch-all@ destinations
               vhost --deluali host_name user_alias|catch-all@ [destinations]
  Manage User: vhost --passwd [host_name] user_name
               vhost --access [host_name] [user_name] 2|1|0
    Set Quota: vhost --qhost host_name mailbox_size(K/M/G) #user #uali
               vhost --quser owner_name disk_space(K/M/G) #host #hali
    List Info: vhost --lshost [host_name|owner_name] [-s|-r[n]] [-l] [-a] [-f]
               vhost --lsuser [host_name|owner_name] [-s|-r[n]]
     View Log: vhost --hostlog host_name [log_name|-q]
               vhost --cmdlog [log_file]
Reload Daemon: vhost --reload [-delay_time(Sec)] [-f]
-c: make Copy instead of symlink [note: "-p profile" can go anywhere in a cmd]
-z,-i: add as dns Zone, add as Ip-based vhost (add reverse dns mapping)
-q,-f,-d: Quit w/o reloading, Force add/del/list/reload, Delete data
-s,-r,-l,-a: Sort, Reverse sort by [n]th field, Long list, list All
#
$HELL = <<'#';
    Auto Conf: vhost --autoconf
   Rotate Log: vhost --rotatelog [host_name] [-z] [-#backlog]
    Get Value: vhost --getvalue [host_name] eval_script
    Uninstall: vhost --uninstall
      Install: vhost --install [-f|-d]
      Qui[e]t: vhost -q
#
use 5;

$INSTALL = $ARGV[0] =~ /^-*install$/;
$ERR = 0;
$ERR = 1, print $SUPPORT if $ARGV[0] =~ /^-*support$/;
$ERR = 1, print $REQUIRE if $ARGV[0] =~ /^-*require$/;
$ERR = 1, print $HISTORY if $ARGV[0] =~ /^-*history$/;
$ERR = 1, print $VERSION if $ARGV[0] =~ /^-*version$/ || $ARGV[0] =~ /^-*v$/;
$ERR && exit 0;
print STDERR $VERSION, "\n" if $INSTALL;

sub Print
{
   print "vhost: $_[0]\n";
}

sub Error
{
   print STDERR "vhost: ", !$_[1] && "error - ", "$_[0]\n";
}

sub Exit
{
   my $err = $_[0] + 0;
   my $host = $HOST && "\@$HOST";
   my @date = split /\s+/, $DATE;
   open FILE, ">> $CMDLOG";
   flock FILE, 2;
   printf FILE "%3s %2s %8s %s %s", @date[1, 2, 3], (split /\./, $HNAM)[0], "vhost[$$:$err]: ($USER$host) --@ARGV -p $PROF\n";
   exit $err;
}

sub Cat
{
   my ($file, @file);
   open FiLe, $_[0];
   @file = <FiLe>;
   close FiLe;
   $file .= $_ for !wantarray && @file;
   return $file || @file;
}

sub Flock
{
   my $ok;
   if ($_[0])
   {
      open FILE, "$_[1] $_[0]";
      $ok = flock FILE, 2;
#     $ok = flock (FILE, 6), $ok && last, sleep 1 for 0..@USER;
      $ERR = 1, Error "unable to lock file '$_[0]'" if !$ok;
      $ERR && Exit 7;
      seek (FILE, 0, 0), @FILE = <FILE> if $_[1] =~ /^\+/;
      seek FILE, 0, $_[2];
      $TELL = tell;
   } else {
      truncate FILE, tell if tell && !$TELL || @FILE && !grep /./s, @FILE;
      close FILE;
      undef @FILE;
   }
}

sub GetFn
{
   my $fn = $_[0];
   my $n = `cd \`dirname "$fn" $NERR\` $NERR; pwd $NERR`; chomp $n;
   $fn = $fn =~ /^\// && $fn || "$n/" . `basename "$fn" $NERR`, chomp $fn if $fn;
   return $fn;
}

sub GetLn
{
   my $ln = $_[0];
   my $n = `cd \`dirname "$ln" $NERR\` $NERR; pwd $NERR`; chomp $n;
   $ln = $ln =~ /^\// && $ln || "$n/" . `basename "$ln" $NERR`, chomp $ln if $ln;
   $ln = readlink $ln, $ln = $ln =~ /^\// && $ln || "$n/" . $ln while -l $ln;
   return $ln;
}

sub GetHa
{
   my ($ha, $i);
   /^\s*$HEAD|^\s*$FOOT|^\s*# (\S+) $SIGN/ && ($ha = $1), 
   /^\s*[^#\s]+\s+\Q$_[0]\E(\s|$)/ && last, $i++ == $#HOST && ($ha = undef) for @HOST;
   return $ha;
}

sub CheckI
{
   $_[0] = $_ if !@_;
   my $err;
   $err = $err || $_[0] !~ /\A\d+\.\d+\.\d+\.\d+\z/ || $_ > 255 || /^0./ for split /\./, $_[0] || 0;
   $err && $_[1] && Error "invalid ip address '$_[0]'";
   return !$err;
}

sub CheckH
{
   $_[0] = $_ if !@_;
   my $err;
   $err = 1, $_[1] && Error "invalid host name '$_[0]'" if $_[0] !~ /\A([A-Z0-9-]+\.)+[A-Z]{2,4}\z/i;
   $err = 1, $_[1] && Error "upper case in host name '$_[0]'" if $_[0] =~ /[A-Z]/ && !$err;
   return !$err;
}

sub CheckU
{
   $_[0] = $_ if !@_;
   my $err;
   $err = 1, $_[1] && Error "invalid user name '$_[0]'" if $_[0] !~ /\A[A-Z][A-Z0-9_.-]*\z/i;
   $err = 1, $_[1] && Error "upper case in user name '$_[0]'" if $_[0] =~ /[A-Z]/ && !$err;
   return !$err;
}

sub CheckF
{
   $_[0] = $_ if !@_;
   my $err;
   $err = 1, $_[1] && Error "invalid file or directory '$_[0]'" if $_[0] !~ /\A[\/\w.-]*[\w.-]\z/ || $_[0] =~ /(^|\/)\.+$/;
   return !$err;
}

sub CheckN
{
   $_[0] = $_ if !@_;
   my $err;
   $err = 1, $_[1] && Error "invalid log name '$_[0]'" if $_[0] !~ /\A[\w.-]+\z/ || $_[0] =~ /^\.+$/;
   return !$err;
}

sub CheckQ
{
   $_[0] = $_ if !@_;
   my $err;
   $err = 1, $_[1] && Error "invalid quota '$_[0]'" if $_[0] !~ /\A\d+(\.\d*)?[KMGT]\s+\d+\s+\d+\z/i;
   return !$err;
}

$HOST = $ENV{VHOST_HOSTNAME} =~ /(.*)/ && $1;
$USER = $ENV{VHOST_USERNAME} =~ /(.*)/ && $1;
$PASS = $ENV{VHOST_PASSWORD} =~ /(.*)/ && $1;
$PROF = $ENV{VHOST_PROFILE} =~ /(.*)/ && $1;
$USRID = $ENV{VHOST_USERID} =~ /(.*)/ && $1;
$LOCAL = $ENV{VHOST_LOCAL} =~ /(.*)/ && $1;
$CHILD = $ENV{VHOST_CHILD} =~ /(.*)/ && $1;
$DEBUG = $ENV{VHOST_DEBUG} =~ /(.*)/ && $1;
$PAGER = $ENV{PAGER} =~ /(.*)/ && $1;
$TERM = $ENV{TERM} =~ /(.*)/ && $1;
$UID = $ENV{UID} =~ /(.*)/ && $1 || $<;
$< = 0, $( = $) = 1999;
undef %ENV;
$ENV{PAGER} = $PAGER, undef $PAGER;
$ENV{TERM} = $TERM, undef $TERM;
$ENV{PATH} = "/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:/usr/local/sbin";
$FreeBSD = `uname 2>/dev/null` =~ /^FreeBSD$/i && "pw ";
umask 0022;

for my $f (split /[#\s,]+/, $REQUIRE)
{
   next if $f =~ /^(|-.*|echo|cd|pwd)$/ || !$INSTALL && $f =~ /^(make|gcc|g\+\+)$/ || 
      ($FreeBSD? $f =~ /^(useradd|usermod|userdel|groupadd|groupmod|groupdel)$/ : $f =~ /^pw$/);
   my $ok;
   $ok = $ok || -f "$_/$f" && -x "$_/$f" for split /:/, $ENV{PATH};
   $ERR = 1, Error "missing system file '$f'" if $f =~ /^\// && !-f $f;
   $ERR = 1, Error "missing system command '$f'" if $f !~ /^\// && !$ok;
}
$ERR && exit 1;

$NERR = "2>/dev/null" if !$DEBUG;
$SIGN = "- added by vhost,";
$FOOT = "# keep this line - by vhost";
$HEAD = "# keep this line, the blank line above, and the line below - by vhost";
$DATE = (`date $NERR` =~ /(.*)/)[0];
$UNIX = (`uname $NERR` =~ /(.*)/)[0];
$HNAM = (`uname -n $NERR` =~ /(.*)/)[0];
$HNAM = lc ((gethostbyname $HNAM)[0] || $HNAM);
$HIP = join ".", unpack "C4", (gethostbyname $HNAM)[4];
$GHID = 1999;
$GUID = 2000;
$ERR = 1, Error "invalid real host name '$HNAM'" if !CheckH ($HNAM) || !CheckI $HIP;
$ERR && exit 1;

if ($HOST || $USER || $PASS)
{
   my $pswd = (getpwnam $USER)[1];
   my $loh = $HOST eq $HNAM || -c "/etc/vdata/vusers.$HOST";
   $USER = $HOST && !$loh && (Cat ("/etc/vdata/vusers.$HOST") =~ /^\s*([^#\s]+)\s+\Q$USER\E(\s|$)/m)[0] || $USER;
   $PSWD = $HOST && !$loh? (Cat ("/etc/vdata/shadow.$HOST") =~ /^\s*\Q$USER\E:([^:]*):/m)[0] : $pswd;
   $USER = undef if (!$PSWD || $PSWD ne crypt $PASS, $PSWD) && !-o (getpwuid $UID)[7]."/.rhosts";
} else {
   $USER = getpwuid $UID;
}
$ERR = 1, Error "login incorrect" if !$USER || !$INSTALL && $> || $INSTALL && ($UID || $>);
$ERR && exit 127;

@HOST = Cat "/etc/vhosts"; /^\s*([^#\s]+)\s+(\S+)\s*(.*)/ && ($HOST{$2} ||= $1, $QHOST{$2} ||= $3) for @HOST;
@USER = Cat "/etc/vusers"; /^\s*([^#\s]+)\s+(\S+)\s*(.*)/ && ($USER{$1} ||= $2, $QUSER{$1} ||= $3) for @USER;
@PROF = </etc/vdata/vhostd.*>; /(\/etc\/vdata\/vhostd\.(.*))/ && ($PROF{$2} ||= readlink $1 || $1) for @PROF;
$HOME{$_} ||= (Cat ("$_/V00hostconf") =~ /^\s*\$VHOME[\s="']+([^\s="';]+)/m)[0] for "/etc/vhost.d", values %USER, values %PROF;
$OWNR{$_} ||= getpwuid ((stat "$HOME{$PROF{$_}}/$_")[4]) || (stat "$HOME{$PROF{$_}}/$_")[4] for keys %PROF;

my $i;
s/\n/ /g, /^-p$/ && ($PROF = $ARGV[$i+1]) && splice (@ARGV, $i, 2) && last, $i++ for @ARGV;
$ARGV[0] =~ s/^-*// if @ARGV;
$ARGV[0] =~ s/^listhost$/lshost/ if @ARGV;
$ARGV[0] =~ s/^listuser$/lsuser/ if @ARGV;
$ERR = 1;
$ERR = 0 if $#ARGV+1 > 1 && $#ARGV-1 < 4 && join ("\n", @ARGV) =~ /^addvroot\n[^-].*(\n[^-].*)?(\n-[cf\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 1 && $#ARGV-1 < 3 && join ("\n", @ARGV) =~ /^delvroot\n[^-].*(\n[^-].*)?(\n-[f\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 1 && $#ARGV-1 < 6 && join ("\n", @ARGV) =~ /^adddns\n[^-].*(\n[^-].*)?(\n-[ziqf\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 1 && $#ARGV-1 < 4 && join ("\n", @ARGV) =~ /^deldns\n[^-].*(\n[^-].*)?(\n-[qf\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 2 && $#ARGV-1 < 7 && join ("\n", @ARGV) =~ /^addhost\n[^-].*\n[^-].*(\n[^-].*)?(\n-[ziqf\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 1 && $#ARGV-1 < 4 && join ("\n", @ARGV) =~ /^delhost\n[^-].*(\n-[qfd\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 2 && $#ARGV-1 < 4 && join ("\n", @ARGV) =~ /^addhali\n[^-].*\n[^-].*(\n-[zq\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 1 && $#ARGV-1 < 2 && join ("\n", @ARGV) =~ /^delhali\n[^-].*(\n-[q\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 2 && $#ARGV-1 < 4 && join ("\n", @ARGV) =~ /^adduser\n[^-].*\n[^-].*(\n[^-].*)?(\n-[f\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 2 && $#ARGV-1 < 4 && join ("\n", @ARGV) =~ /^deluser\n[^-].*\n[^-].*(\n-[fd\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 3 && $#ARGV-1 > 0 && join ("\n", @ARGV) =~ /^adduali\n[^-].*\n[^-].*(\n.+)+$/;
$ERR = 0 if $#ARGV+1 > 2 && $#ARGV-1 > 0 && join ("\n", @ARGV) =~ /^deluali\n[^-].*\n[^-].*(\n.+)*$/;
$ERR = 0 if $#ARGV+1 > 1 && $#ARGV-1 < 3 && join ("\n", @ARGV) =~ /^passwd(\n[^-].*)?\n[^-].*(\n.+)?$/;
$ERR = 0 if $#ARGV+1 > 2 && $#ARGV-1 < 4 && join ("\n", @ARGV) =~ /^access(\n[^-].*)?(\n[^-].*)?\n[210](\n-[hu\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 4 && $#ARGV-1 < 4 && join ("\n", @ARGV) =~ /^qhost\n[^-].*\n[^-].*\n[^-].*\n[^-].*$/;
$ERR = 0 if $#ARGV+1 > 4 && $#ARGV-1 < 4 && join ("\n", @ARGV) =~ /^quser\n[^-].*\n[^-].*\n[^-].*\n[^-].*$/;
$ERR = 0 if $#ARGV+1 > 0 && $#ARGV-1 < 6 && join ("\n", @ARGV) =~ /^lshost(\n[^-].*)?(\n-[srlafhu\d\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 0 && $#ARGV-1 < 3 && join ("\n", @ARGV) =~ /^lsuser(\n[^-].*)?(\n-[srhu\d\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 1 && $#ARGV-1 < 2 && join ("\n", @ARGV) =~ /^hostlog\n[^-].*(\n[^-].*)?(\n-[q\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 0 && $#ARGV-1 < 1 && join ("\n", @ARGV) =~ /^cmdlog(\n[^-].*)?$/;
$ERR = 0 if $#ARGV+1 > 0 && $#ARGV-1 < 3 && join ("\n", @ARGV) =~ /^reload(\n[\d]+)?(\n-[lf\d\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 0 && $#ARGV-1 < 0 && join ("\n", @ARGV) =~ /^autoconf$/;
$ERR = 0 if $#ARGV+1 > 0 && $#ARGV-1 < 3 && join ("\n", @ARGV) =~ /^rotatelog(\n[^-].*)?(\n[\d]+)?(\n-[z\d\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 1 && $#ARGV-1 < 2 && join ("\n", @ARGV) =~ /^getvalue(\n[^-].*)?\n.+$/;
$ERR = 0 if $#ARGV+1 > 0 && $#ARGV-1 < 0 && join ("\n", @ARGV) =~ /^uninstall$/;
$ERR = 0 if $#ARGV+1 > 0 && $#ARGV-1 < 1 && join ("\n", @ARGV) =~ /^install(\n-[fd\s-]+)?$/;
$ERR = 0 if $#ARGV+1 > 0 && $#ARGV-1 < 0 && join ("\n", @ARGV) =~ /^q$/;
$ERR = 0 if $HLP = "$HELP$HELL" !~ / vhost -+$ARGV[0]\s/;
$ERR = $ERR || !length for @ARGV;
$ERR && print STDERR $HELP;
$ERR && exit 2;

$_c = $_c || /^-.*c/ && "-c" for @ARGV;
$_z = $_z || /^-.*z/ && "-z" for @ARGV;
$_i = $_i || /^-.*i/ && "-i" for @ARGV;
$_q = $_q || /^-.*q/ && "-q" for @ARGV;
$_f = $_f || /^-.*f/ && "-f" for @ARGV;
$_d = $_d || /^-.*d/ && "-d" for @ARGV;
$_s = $_s || /^-.*s/ && "-s" for @ARGV;
$_r = $_r || /^-.*r/ && "-r" for @ARGV;
$_l = $_l || /^-.*l/ && "-l" for @ARGV;
$_a = $_a || /^-.*a/ && "-a" for @ARGV;
$_h = $_h || /^-.*h/ && "-h" for @ARGV;
$_u = $_u || /^-.*u/ && "-u" for @ARGV;
$_n = $_n || /^-\D*(\d+)|^(\d+)$/ && ($1 || $2) for @ARGV;

if ($ARGV[0] eq "addvroot")
{
   $P = $ARGV[1];
   $W = $ARGV[2] !~ /^-/ && $ARGV[2];
}
if ($ARGV[0] eq "delvroot")
{
   $P = $ARGV[1];
   $W = $ARGV[2] !~ /^-/ && $ARGV[2];
}
if ($ARGV[0] eq "adddns")
{
   $H = $ARGV[1];
   $I = $ARGV[2] !~ /^-/ && $ARGV[2] || ".";
}
if ($ARGV[0] eq "deldns")
{
   $H = $ARGV[1];
   $I = $ARGV[2] !~ /^-/ && $ARGV[2];
}
if ($ARGV[0] eq "addhost")
{
   $H = $ARGV[1];
   $O = $ARGV[2];
   $I = $ARGV[3] !~ /^-/ && $ARGV[3] || ".";
}
if ($ARGV[0] eq "delhost")
{
   $H = $ARGV[1];
}
if ($ARGV[0] eq "addhali")
{
   $H = $ARGV[1];
   $A = $ARGV[2];
}
if ($ARGV[0] eq "delhali")
{
   $A = $ARGV[1];
   $H = GetHa $A;
}
if ($ARGV[0] eq "adduser")
{
   $H = $ARGV[1];
   $U = $ARGV[2];
   $N = $ARGV[3] !~ /^-/ && $ARGV[3];
}
if ($ARGV[0] eq "deluser")
{
   $H = $ARGV[1];
   $U = $ARGV[2];
}
if ($ARGV[0] eq "adduali")
{
   $H = $ARGV[1];
   $V = $ARGV[2];
   @N = @ARGV[3..$#ARGV];
}
if ($ARGV[0] eq "deluali")
{
   $H = $ARGV[1];
   $V = $ARGV[2];
   @N = @ARGV[3..$#ARGV];
}
if ($ARGV[0] eq "passwd")
{
   $H = $#ARGV > 1? $ARGV[1] : undef;
   $U = $#ARGV > 1? $ARGV[2] : $ARGV[1];
   $N = $ARGV[3];
   $O = $U if !$H;
}
if ($ARGV[0] eq "access")
{
   $H = $ARGV[-1] !~ /^-/ && $#ARGV > 2 || $#ARGV > 3? $ARGV[1] : ($_h || $ARGV[1] =~ /\./) && $ARGV[1];
   $U = $ARGV[-1] !~ /^-/ && $#ARGV > 2 || $#ARGV > 3? $ARGV[2] : ($_u || $ARGV[1] !~ /\./) && $ARGV[1];
   $S = $ARGV[-1] !~ /^-/ && $ARGV[-1] || $ARGV[-2];
   $H = undef if $_h && $_u;
   $O = $U if !$H;
}
if ($ARGV[0] eq "qhost")
{
   $H = $ARGV[1];
   $Q = "@ARGV[2..4]";
}
if ($ARGV[0] eq "quser")
{
   $O = $ARGV[1];
   $Q = "@ARGV[2..4]";
}
if ($ARGV[0] eq "lshost")
{
   $H = $ARGV[1] !~ /^-/ && ($_h || $ARGV[1] =~ /\./) && $ARGV[1];
   $O = $ARGV[1] !~ /^-/ && ($_u || $ARGV[1] !~ /\./) && $ARGV[1];
   $H = undef if $_h && $_u;
}
if ($ARGV[0] eq "lsuser")
{
   $H = $ARGV[1] !~ /^-/ && ($_h || $ARGV[1] =~ /\./) && $ARGV[1];
   $O = $ARGV[1] !~ /^-/ && ($_u || $ARGV[1] !~ /\./) && $ARGV[1];
   $O = undef if $_h && $_u;
}
if ($ARGV[0] eq "hostlog")
{
   $H = $ARGV[1];
   $L = $ARGV[2] !~ /^-/ && $ARGV[2];
}
if ($ARGV[0] eq "cmdlog")
{
   $G = $ARGV[1];
}
if ($ARGV[0] eq "rotatelog")
{
   $H = $ARGV[1] !~ /^-|^\d+$/ && $ARGV[1];
}
if ($ARGV[0] eq "getvalue")
{
   $H = $#ARGV > 1 && $ARGV[1];
   $Y = $ARGV[-1];
}
if ($ARGV[0] eq "vdeliver")
{
   $H = lc $ARGV[-1];
   $H = GetHa ($H) || $H;
}

$ERR = !CheckH ($H, 1) || $ERR if length $H;
$ERR = !CheckH ($A, 1) || $ERR if length $A;
$ERR = !CheckU ($U, 1) || $ERR if length $U;
$ERR = !CheckU ($V, 1) || $ERR if length $V && $V ne "catch-all@";
$ERR = !CheckU ($O, 1) || $ERR if length $O;
$ERR = !CheckU ($W, 1) || $ERR if length $W;
$ERR = !CheckI ($I, 1) || $ERR if length $I && $I ne ".";
$ERR = !CheckF ($P, 1) || $ERR if length $P;
$ERR = !CheckF ($G, 1) || $ERR if length $G;
$ERR = !CheckN ($L, 1) || $ERR if length $L;
$ERR = !CheckQ ($Q, 1) || $ERR if length $Q;
$ERR && exit 2;

if ($INSTALL)
{
   $< = (stat "./vhost.pl")[4];
   $ERR = 1, Error "invalid or insecure install module" if !-f "./vmake" || !-o "./vmake" && !-O "./vmake";
   $ERR && exit 1;
   $< = 0;
   do "./vmake";
}
$< = $UID;
$PROF = -f "$P/V00hostconf" && $P || CheckQ ($QHOST{$H}) && $PROF{$H} || $PROF || $USER{$O} || !$HOST && $USER{$USER} || 
   "/etc/vhost.d";
$ERR = 1, Error "invalid or insecure profile '$PROF'" if !-d $PROF || !-o $PROF || -W $PROF && $UID || !CheckF $PROF;
$ERR && exit 1;
$ERR = 1, Error "missing '$PROF/V00hostconf'" if !-f "$PROF/V00hostconf";
$ERR && exit 1;
$< = 0;

$PROF = GetFn $PROF;
require for <$PROF/V[0-9][0-9]*>;
sub Do
{
   $X = $_[0];
   $MOD = "$PROF/V00hostconf", eval V00hostconf;
   $MOD = $_, /\/V00hostconf$/ || do $_ for <$PROF/V[0-9][0-9]*>;
}
Do "configuration";
$ERR && Error "please adjust system configuration(s)", 1;
$ERR && exit 1;
Do "initialization";
Do $ARGV[0];
Do "finalization";
$HLP && print STDERR $HELP;
$HLP && exit 2;

Exit $ERR;


sub V00hostconf { ############################################################
$X eq "configuration" && do { ################################################

require Quota;
require IO::Pty;
do "$PROF/V00hostconf";
$ARGV[0] =~ /(^reload|dns)$/ && ($LOCAL ||= !$HOSTs && !$DNSs);
$ARGV[0] !~ /(^reload|dns)$/ && ($LOCAL ||= !$HOSTs);
my $err;
$err = 0; $err ||= !CheckH && !CheckI for @HOSTs = split /\s+/, $HOSTs;
$ERR = 1, Error "$MOD: \$HOSTs = '$HOSTs'" if $err;
$err = 0; $err ||= !CheckH && !CheckI for @DNSs = split /\s+/, $DNSs;
$ERR = 1, Error "$MOD: \$DNSs = '$DNSs'" if $err;
$err = 0; $err ||= !CheckI for @IPs = split /-|\s+/, $IPs;
$ERR = 1, Error "$MOD: \$IPs = '$IPs'" if $err;
$err = 0; $err ||= !CheckH && $_ for @TLDs = split /^\.*|\.*\s+\.*|\.*$/, $TLDs;
$ERR = 1, Error "$MOD: \$TLDs = '$TLDs'" if $err;
$ERR = 1, Error "$MOD: \$HOSTADD = '$HOSTADD'" if $HOSTADD !~ /\A[210]\z/;
$ERR = 1, Error "$MOD: \$IPADD = '$IPADD'" if $IPADD !~ /\A[10]\z/;
$ERR = 1, Error "$MOD: \$DNSAWARE = '$DNSAWARE'" if $DNSAWARE !~ /\A[10]\z/;
$ERR = 1, Error "$MOD: \$UIDSYNC = '$UIDSYNC'" if $UIDSYNC !~ /\A\d+\z/;
$ERR = 1, Error "$MOD: \$TIMEOUT = '$TIMEOUT'" if $TIMEOUT !~ /\A\d+\z/;
$ERR = 1, Error "$MOD: \$ROOTCMD = '$ROOTCMD'" if $ROOTCMD !~ /\A[10]\z/;
$ERR = 1, Error "$MOD: \$RUSER = '$RUSER'" if $RUSER eq "root" || !CheckU $RUSER;
$ERR = 1, Error "$MOD: \$REXEC = '$REXEC'" if $REXEC !~ /\s+-l\s+\Q$RUSER\E\s*\z/;
$ERR = 1, Error "$MOD: \$VHOME = '$VHOME'" if $VHOME !~ /^\// || !CheckF $VHOME;
$ERR = 1, Error "$MOD: \$HTEMP = '$HTEMP'" if $HTEMP && ($HTEMP !~ /^\// || !CheckF $HTEMP);
$ERR = 1, Error "$MOD: \$UTEMP = '$UTEMP'" if $UTEMP && ($UTEMP !~ /^\// || !CheckF $UTEMP);
$ERR = 1, Error "$MOD: \$ADMIN = '$ADMIN'" if !CheckU $ADMIN;
$ERR = 1, Error "$MOD: \$SHELL = '$SHELL'" if !grep /^\s*\Q$SHELL\E\s*$/, Cat "/etc/shells";
$ERR = 1, Error "$MOD: \$ACCESS = '$ACCESS'" if $ACCESS !~ /\A[210]\z/;
$ERR = 1, Error "$MOD: \$OWNERMAX = '$OWNERMAX'" if $OWNERMAX !~ /\A\d+\z/;
$ERR = 1, Error "$MOD: \$QHOST = '$QHOST'" if !CheckQ $QHOST;
$ERR = 1, Error "$MOD: \$QUSER = '$QUSER'" if !CheckQ $QUSER;
$ERR = 1, Error "$MOD: \$CMDLOG = '$CMDLOG'" if $CMDLOG !~ /^\// || !CheckF $CMDLOG;
$ERR = 1, Error "$MOD: \$PAGER = '$PAGER'" if $PAGER !~ /^\// || !-x (split /[\s|&;]+/, $PAGER)[0];
$ERR = 1, Error "$MOD: \$DUCACHE = '$DUCACHE'" if $DUCACHE !~ /\A[10]\z/;
$ERR = 1, Error "$MOD: \$AUTOCONF = '$AUTOCONF'" if $AUTOCONF !~ /\A[10]\z/;
$ERR = 1, Error "$MOD: \$dotChar = '$dotChar'" if $dotChar !~ /\A[\w.-]\z/;
$ERR = 1, Error "$MOD: \$unameLen = '$unameLen'" if !$unameLen || $unameLen !~ /\A\d+\z/;
$ERR = 1, Error "$MOD: \$minLogin = '$minLogin'" if !$minLogin || $minLogin !~ /\A\d+\z/;
$ERR = 1, Error "$MOD: \$blockSize = '$blockSize'" if !$blockSize || $blockSize !~ /\A\d+\z/;
$ERR = 1, Error "$MOD: \$homeDir = '$homeDir'" if !-d $homeDir;
$ERR = 1, Error "$MOD: \$mailDir = '$mailDir'" if !-d $mailDir;
$ERR = 1, Error "$MOD: \$cronDir = '$cronDir'" if !-d $cronDir;
$ERR = 1, Error "$MOD: \$passwdCf = '$passwdCf'" if !-f $passwdCf;
$ERR = 1, Error "$MOD: \$shadowCf = '$shadowCf'" if !-f $shadowCf;
$ERR = 1, Error "$MOD: \$aliasesCf = '$aliasesCf'" if !-f $aliasesCf;
$ERR = 1, Error "$MOD: \$inetdCf = '$inetdCf'" if !-e $inetdCf;
$ERR = 1, Error "$MOD: \$inetdRc = '$inetdRc'" if $inetdRc && ($inetdRc !~ /^\// || !-x (split /[\s|&;]+/, $inetdRc)[0]);
$ARGV[0] =~ /dns$/ || next;
$RDNS = $DNSs[0] if !$LOCAL && $DNSs !~ /(^|\s)(\Q$HNAM\E|\Q$HIP\E)(\s|$)/;
$RDNS = "$REXEC $RDNS" if $RDNS;
@dnsCf = `$RDNS cat $dnsCf $NERR` if -f $dnsCf;
@revZf = `$RDNS cat $revZf $NERR` if -f $revZf;
$ERR = 1, Error "$MOD: \$dnsCf = '$dnsCf'" if !@dnsCf;
$ERR = 1, Error "$MOD: \$dnsRc = '$dnsRc'" if $dnsRc && ($dnsRc !~ /^\// || !-x (split /[\s|&;]+/, $dnsRc)[0]);
$ERR = 1, Error "$MOD: \$revZf = '$revZf'" if !@revZf && $_i;


};
$X eq "initialization" && do { ###############################################

sub Error5
{
   $ERR = 1, Error $_[0];
   $ERR && Exit 5;
}

sub Prompt
{
   my $yn;
   print "vhost: $_[0]. continue? y/n[n]: ";
   chomp ($yn = <STDIN>) if !$_f;
#  print "$yn\n" if !$_f && $ZTERM;
   print $yn = "y\n" if $_f || $ZTERM;
   $RECHO .= "$yn\n" if !$_f;
   $yn !~ /^y/i && Exit 6;
}

sub Forced
{
   my $yn;
   print "vhost: $_[0]? y/n[n]: ";
   chomp ($yn = <STDIN>) if !$_f;
   print "$yn\n" if !$_f && $ZTERM;
   print $_d? "y\n" : "n\n" if $_f;
   $RECHO .= "$yn\n" if !$_f;
   $_d = $yn =~ /^y/i && "-d" if !$_f;
}

sub Passwd
{
   my ($pass, $pswd);
   print "vhost: $_[0]: ";
   chomp ($pass = <STDIN>) if !$_f;
   print "$pass\n" if !$_f && $ZTERM;
   print "\n" if $_f;
   $pswd = $pass;
   $pswd =~ s/([\\\$"`])/\\\\\\$1/g;
   $RECHO .= "$pswd\n" if !$_f;
   $ERR = 1, Error "no password provided" if !$_f && !$pass;
   $ERR && Exit 6;
   return $pass;
}

sub Crypt
{
   my ($salt, $pswd);
   $salt .= (".", "/", 0..9, "A".."Z", "a".."z")[rand 64] for 1..8;
   $pswd = crypt $_[0], "\$1\$$salt";
   return $pswd;
}

sub SetPass
{
   my ($i, $tmp);
   my $master = new IO::Pty;
   my $slave = $master->slave;
   if (fork == 0 && $master->make_slave_controlling_terminal)
   {
      close $master;
      open STDIN,  "<&".$slave->fileno;
      open STDOUT, ">&".$slave->fileno;
      open STDERR, ">&".$slave->fileno;
      exec "passwd $_[0]; echo $SIGN" if length $_[0] && length $_[1];
      close $slave;
      exit 0;
   }
   close $slave;
   sleep 1 && print ($master "$_[1]\n") while sleep 1 && sysread ($master, $tmp, 99999) && $tmp !~ /$SIGN/ && $i++ < 5;
   close $master;
}

sub Rexec
{
   my ($env, @his);
   $env .= "VHOST_HOSTNAME=$HOST';' export VHOST_HOSTNAME';' ";
   $env .= "VHOST_USERNAME=$USER';' export VHOST_USERNAME';' ";
   $env .= "VHOST_USERID=$USRID';' export VHOST_USERID';' ";
   $env .= "VHOST_LOCAL=$LOCAL';' export VHOST_LOCAL';' ";
   $env .= "VHOST_CHILD=$CHILD';' export VHOST_CHILD';' ";
   my $cmd = $_[0];
   $cmd =~ s/\s+/ /g;
   $RECHO .= "y\ny\ny\n";
   $_ && "@his" !~ /(^|\s)\Q$_\E(\s|$)/ && ($his[$#his+1] = $_) for split /\s+/, $_[2];
   $< = $RID;
   print ("[$_[1] $_]\n$cmd\n"), 
   system ("$REXEC $_ $env echo \"\\\"$RECHO\\\"\" '|' /usr/sbin/$_[0] -p $PROF 2>&1") for @his;
   Exit 0;
}

sub Rdns
{
   my $env;
   $env .= "VHOST_USERNAME=$USER; export VHOST_USERNAME; ";
   $env .= "VHOST_LOCAL=; export VHOST_LOCAL; ";
   $< = $RID;
   system "$env /usr/sbin/$_[0] -p $PROF 2>&1";
   $< = 0;
}

sub GroupAdd
{
   my @ghid = split /\D+/, `id -G $_[0] $NERR`;
   my $ghid = join ",", $GHID, @ghid if !grep /^\Q$GHID\E$/, @ghid;
   `${FreeBSD}usermod -G $ghid $_[0] $NERR` if $ghid;
}

sub BackupCf
{
   `cp -p $_[0] $_[0].vhost.0 $NERR` if -f FILE && !-s "$_[0].vhost.0";
}

sub AddConf
{
   my ($i, $j);
   my $conf = $_[1];
   $conf =~ s/\[VROOT_NAME\]/$W/g;
   $conf =~ s/\[PROFILE_DIR\]/$P/g;
   $conf =~ s/\[ZONE_NAME\]/$Z/g;
   $conf =~ s/\[MASTER_IP\]/$MASIP/g;
   $conf =~ s/\[HOST_NAME\]/$H/g;
   $conf =~ s/\[HOST_ALIAS\]/$A/g;
   $conf =~ s/\[OWNER_NAME\]/$O/g;
   $conf =~ s/\[IP_ADDRESS\]/$I/g;
   $conf =~ s/\[USER_NAME\]/$U/g;
   $conf =~ s/\[USER_ALIAS\]/$V/g;
   Flock $_[0], "+<", 0 if $HOME{$PROF{$H}} && $HOME{$PROF{$H}} ne $VHOME || $HOST{$H} && $I && $HOST{$H} ne $I;
   /^\s*# \Q$H\E $SIGN/ && ($j = $i), $j && /^\s*$FOOT/ && last, 
   $j && s/(^|\s)\Q$HOME{$PROF{$H}}\E(\/\Q$H\E)/$1$VHOME$2/g, $j && s/(^|\D)\Q$HOST{$H}\E(\D|$)/$1$I$2/g, $i++ for @FILE;
   print FILE @FILE if $i <= $#FILE;
   Flock 0;
   Flock $_[0], "+>>", 2;
   print FILE "\n$HEAD\n# $H $SIGN $DATE\n$conf$FOOT\n" if $conf && !grep /^\s*# \Q$H\E $SIGN/, @FILE;
   Flock 0;
}

sub DelConf
{
   my ($i, $j);
   Flock $_[0], "+<", 0;
   s/^\s*# \Q$H\E $SIGN.*//s && ($j = $i), $j && s/^\s*$FOOT.*//s && last, $j && ($_ = ""), $i++ for @FILE;
   print FILE @FILE if $i <= $#FILE && $FILE[$j-1] =~ s/^\s*$HEAD.*//s && $FILE[$j-2] =~ s/^\s*$//s;
   Flock 0;
}

sub ReloadRc
{
   my $rc = $_[0];
   my $rf = $MOD;
   $rc =~ s/(;|$)/ 2>&1$1/g if $rc;
   $rf =~ s/.*\/V(\d\d)([^\/]*)$/r$1.$2/;
   `touch /tmp/.vhost/.$rf $NERR` if $rc eq "0";
   $_ && chomp && Print $_ for -f "/tmp/.vhost/.$rf" && $rc && `$rc`;
   `rm /tmp/.vhost/.$rf $NERR` if -f "/tmp/.vhost/.$rf" && $rc ne "0";
}

sub UpdateFs
{
   my ($zf, $ok);
   my @fs = localtime;
   my $fs = ($fs[5] + 1900) * 1000000 + ($fs[4] + 1) * 10000 + $fs[3] * 100;
   my $file = ($zf = (split /\//, $_[0])[-1]) && [Cat "/tmp/.vhost/.$zf"] || \@FILE;
   `rm /tmp/.vhost/.$zf $NERR`;
   $ok = $ok || /\sin\s+soa\s/i, 
   $ok && /(^|\s)(\d+)(\s|$)/ && ($fs = $fs>$2? $fs : $2+1) && s/(^|\s)(\d+)(\s|$)/$1$fs$3/ && last for @{$file};
   return $fs;
}

sub GetZf
{
   my ($zf, $ok, $i, @a, @z);
   $ok = $ok || /^\s*zone[\s"']+\Q$_[0]\E[\s"']/i, @a = /\{/g, @z = /\}/g, $i += $#a - $#z, 
   $ok && (/\bfile[\s"']+([^\s"';]+)/i && ($zf = $1) || /\}/ && $i <= 0) && last for @dnsCf;
   my ($zd, $ok, $i, @a, @z);
   $ok = $ok || /^\s*options\b/i, @a = /\{/g, @z = /\}/g, $i += $#a - $#z, 
   $ok && (/\bdirectory[\s"']+([^\s"';]+)/i && ($zd = $1) || /\}/ && $i <= 0) && last for @dnsCf;
   return "$zd/$zf" if $zf;
}

sub GetZ
{
   my $z;
   my @z = split /\./, $_[0];
   $z = join (".", @z[$_..$#z]), ( 
      $ARGV[0] =~ /dns$/ && ($_z || grep /^\s*zone[\s"']+\Q$z\E[\s"']/i, @dnsCf) || 
      $ARGV[0] !~ /dns$/ && $PROF{$z} || grep $z =~ /^[^.]+\.\Q$_\E$/, @TLDs 
   ) && last for 0..$#z-1;
   return $z;
}

sub GetI
{
   sub i2n
   {
      my @i = split /\./, $_[0];
      my $n = ($i[0] << 24) + ($i[1] << 16) + ($i[2] << 8) + $i[3];
      return $n;
   }
   sub n2i
   {
      my $n = $_[0];
      my $i = join ".", $n >> 24, $n << 8 >> 24, $n << 16 >> 24, $n << 24 >> 24;
      return $i;
   }
   my ($n, %i);
   my $i = $HOST{$H} || $HOST{GetZ $H};
   $i = $i || $IPs[0] eq (split /\s+/, $IPs)[0] && $IPs[0] || $HIP if $IPADD == 0;
   $i = $i || $HIP if !$IPs || $IPADD == 3;
   for (split /\s+/, $IPs)
   {
      my $ip;
      my $n1 = i2n ((split /-/)[0]);
      my $n2 = i2n ((split /-/)[1]) || $n1;
      $ip = n2i (($n1<$n2?$n1:$n2) + $_), $I[$#I+1] = $ip, 
      $IPADD == 1 && ($i{$ip} = grep $_ eq $ip, values %HOST), 
      $IPADD == 2 && () for 0 .. abs $n1 - $n2;
   }
   while (!$i && %i && $IPADD == 1)
   {
      $i{$_} == $n && ($i = $_) && last for @I;
      $n++;
   }
   while (!$i && %i && $IPADD == 2)
   {
      $i = "?";
   }
   return $i;
}

sub GetU
{
   `touch /tmp/.vhost/.uid $NERR` if !-e "/tmp/.vhost/.uid";
   if (!$_[0])
   {
      my ($ok, $loh, $env, $uid, @uid);
      Flock "/tmp/.vhost/.uid", "+<", 0;
      $< = $RID;
      for my $h (split /\s+/, $HOSTs)
      {
         $loh = $h eq $HNAM || $h eq $HIP || grep /(^|\s)\Q$h\E(\s|$)/, @HOST;
         $env .= "VHOST_USERNAME=$USER';' export VHOST_USERNAME';' ";
         $env .= "VHOST_LOCAL=$loh';' export VHOST_LOCAL';' ";
         $uid[$_] = 1 for `$REXEC $h $env /usr/sbin/vhost --getvalue \\\"GetU 1\\\;\\\" -p $PROF $NERR`;
      }
      $< = 0;
      $uid[$_] = 1, $ok = $ok || $_ for time - (stat "/tmp/.vhost/.uid")[9] < 300 && @FILE;
      $uid++ while $uid[$uid] || $uid < ($UIDSYNC < 60000? $UIDSYNC : 60000);
      print FILE @FILE if $ok;
      print FILE "$uid\n";
      Flock 0;
      return $uid;
   } else {
      my ($i, @u, @uid);
      Flock "/tmp/.vhost/.uid", "+<", 2 if !$LOCAL;
      $uid[$i++] = $u[2] while @u = getpwent;
      $uid[$i++] = $_ for time - (stat "/tmp/.vhost/.uid")[9] < 300 && @FILE;
      Flock 0;
      return @uid;
   }
}

sub GetAc
{
   my $ok;
   my $ac = 0;
   $ok = $ok || /^\s*# (\S+) $SIGN/ && $OWNR{$1} eq $_[0] && ($ac-- || 1), 
   $ok = $ok && !/^\s*$HEAD|^\s*$FOOT/, $ok && /^\s*[^#\s]+\s/ && $ac++ for @HOST;
   return $ac;
}

sub GetMp
{
   my $mp;
   my $p = $_[0];
   my @p = split /\//, $p;
   @p = split /\//, $p while $p = readlink $p;
   $mp = join ("/", @p[0..$#p-$_]) || "/", grep (/^\s*[^#\s]+\s+\Q$mp\E\s/, Cat "/etc/fstab") && last for 0..$#p;
   return $mp;
}

sub GetSh
{
   my $sh;
   my $h = $_[1]? $VSPW{$_[0]} !~ /^!*\Q$SHELL[0]\E/ : (getpwnam $_[0])[8];
   $sh = "2" if $_[1]? 0 : 1;
   $sh = "1" if $_[1]? $h : $h eq $SHELL[1];
   $sh = "0" if $_[1]? !$h : $h eq $SHELL[0] || !$h;
   $sh = "-" if $_[1]? 0 : !(getpwnam $_[0])[0];
   return $sh;
}

sub GetKb
{
   my $kb;
   my $b = int $_[0] =~ /^\// && `du -sk $_[0] $NERR` * 1024 || $_[0] > 0 && $_[0];
   my $l = length $b;
   $b = sprintf ("%.".($l%3?2:3)."f", $b / ("1"."0"x$l)) * ("1"."0"x$l);
   $kb = sprintf "%.3sT", $b / 1e12 . ".0" if $b >= 1e12;
   $kb = sprintf "%.3sG", $b / 1e9  . ".0" if $b >= 1e9 && !$kb;
   $kb = sprintf "%.3sM", $b / 1e6  . ".0" if $b >= 1e6 && !$kb;
   $kb = sprintf "%.3sK", $b / 1e3  . ".0" if !$kb;
   return $kb;
}

sub GetNb
{
   my $nb;
   my $b = substr $_[0], -1;
   my %b = ("T", 1e12, "G", 1e9, "M", 1e6, "K", 1e3);
   $b =~ y/kmgt/KMGT/;
   $nb = $_[0] * $b{$b} - !$b{$b};
   return $nb;
}

sub CleanO
{
   my $h;
   my $p = $P || $PROF{$H};
   my $o = $_[0];
   my @o = getpwnam $o;
   my @user = Cat "/etc/vusers";
   my %prof = %PROF;
   my %ownr = %OWNR;
   $prof{$H} = undef;
   $ownr{$H} = undef;
   if (!grep $prof{$_} eq $p && $ownr{$_} eq $o, keys %prof)
   {
      Flock "/etc/vusers", "+<", 0;
      s/^\s*\Q$o\E\s+\Q$p\E\s.*\d\s*$//s for @FILE;
      print FILE @FILE;
      Flock 0;
   }
   if (!grep (/^\s*\Q$o\E\s.*\svroot\s*$/, @user) && $o[3] == $GUID && grep (/^\Q$o\E$/, values %ownr) == 1)
   {
      $ownr{$_} eq $o && ($h = $_) && last for keys %prof;
      `${FreeBSD}usermod -d $o[7]/$h $o $NERR` if $h && $o[7] eq $HOME{$USER{$o}};
   }
   if (!grep (/^\s*\Q$o\E\s.*\svroot\s*$/, @user) && $o[3] == $GUID && grep (/^\Q$o\E$/, values %ownr) == 0)
   {
      Flock "/etc/vusers", "+<", 0;
      s/^\s*\Q$o\E\s.*//s for @FILE;
      print FILE @FILE;
      Flock 0;
      `rm $mailDir/$o $NERR`;
      `${FreeBSD}userdel $o $NERR`;
   }
}

sub AllowO
{
   my $ok;
   my $o = $_[0];
   my $w = $_[1] || $USER;
   my $p = $_[2] || '\S+';
   $p =~ s/\./\\./g;
   my @user = grep /^\s*\Q$w\E\s+$p\s.*\svroot\s*$/, @USER;
   $ok = grep /^\s*\Q$o\E\s+(\S+)/ && grep (/^\s*\Q$w\E\s+\Q$1\E\s.*\svroot\s*$/, @user), @USER;
   $ok = $ok || $o eq $w || $w eq "root";
   $ok = $ok && length $o;
#  $ok || Error "not allowed to access host owner '$o'";
   return $ok;
}

sub AllowI
{
   my $ok;
   my $i = $_[0];
   for (split /\s+/, $IPs)
   {
      my @ip = (split /\./, $i)[0..3];
      my @i1 = (split /[.-]/)[0..3];
      my @i2 = ((split /[.-]/)[4..7], @i1);
      $ok = 1;
      $ok = $ok && ($ip[$_] >= $i1[$_] && $ip[$_] <= $i2[$_] || $ip[$_] >= $i2[$_] && $ip[$_] <= $i1[$_]) for 0..3;
      $ok && last;
   }
   $ok = $ok || !length $IPs;
   $ok = $ok && length $i;
   $ok || Error "not allowed to access ip address '$i'";
   return $ok;
}

sub AllowR
{
   my ($err, $end, $key, $HOSTs_, $DNSs_, $RUSER_);
   my $re = "HOSTs|DNSs|RUSER";
   my $r = $_[0] || $RUSER;
   for my $d (grep !-l && -f "$_/V00hostconf", keys %HOME)
   {
      /^\s*\$($re)\s*=\s*(<<\s*)?["'](.*)["']\s*;\s*$/, $2 && ($end = $3, $key = $1), $1 && eval "\$$1_ = !'$2' && \"$3\";", 
      /^\Q$end\E/ && $end && ($end = ""), !/\Q$end\E/ && $end && eval "\$${key}_ .= \"$_\";", 
      /^\s*### DO NOT TOUCH ANYTHING/ && last for Cat "$d/V00hostconf";
      $err = 1, last if $RUSER_ eq $r && ($HOSTs_ ne $HOSTs || $DNSs_ ne $DNSs);
   }
   return !$err;
}

sub Quota
{
   if ($ARGV[0] eq "addhost")
   {
      my $count = grep /^\s*[^#\s]+\s+\Q$PROF\E\s/, @USER;
      $ERR = 1, Error "$USER: quota exceeded - host owner max $OWNERMAX" if $count >= $OWNERMAX && $OWNERMAX && !@O;
      $ERR && Exit 4;
      my $count = grep $OWNR{$_} eq $O, keys %PROF;
      my $quota = (split /\s+/, $QUSER{$O})[1];
      $ERR = 1, Error "$O: quota exceeded - host max $quota" if $count >= $quota && $quota && !$PROF{$H};
      $ERR && Exit 4;
   }
   if ($ARGV[0] eq "addhali")
   {
      my $count = GetAc $O;
      my $quota = (split /\s+/, $QUSER{$O})[2];
      $ERR = 1, Error "$O: quota exceeded - host alias max $quota" if $count >= $quota && $quota && !$HOST{$H};
      $ERR && Exit 4;
   }
   if ($ARGV[0] eq "adduser")
   {
      my $count = grep /^\s*[^#:]+:/, @VPWD;
      my $quota = (split /\s+/, $QHOST{$H})[1];
      $ERR = 1, Error "$H: quota exceeded - user max $quota" if $count >= $quota && $quota && !$VPWD{$U};
      $ERR && Exit 4;
   }
   if ($ARGV[0] eq "adduali")
   {
      my $count = grep /^\s*[^#:]+:/, @VALI;
      my $quota = (split /\s+/, $QHOST{$H})[2];
      $ERR = 1, Error "$H: quota exceeded - user alias max $quota" if $count >= $quota && $quota && !$VALI{$V};
      $ERR && Exit 4;
   }
   if ($ARGV[0] eq "qhost")
   {
      my @q = split /\s+/, $QHOST;
      $ERR = 1, Error "quota limit (@q) exceeded" 
         if (GetNb ($Q[0]) > GetNb ($q[0]) || !GetNb ($Q[0])) && GetNb ($q[0]) || 
            ($Q[1] > $q[1] || !$Q[1]) && $q[1] || ($Q[2] > $q[2] || !$Q[2]) && $q[2];
      $ERR && Exit 4;
   }
   if ($ARGV[0] eq "quser")
   {
      my @q = split /\s+/, $QUSER;
      $ERR = 1, Error "quota limit (@q) exceeded" 
         if (GetNb ($Q[0]) > GetNb ($q[0]) || !GetNb ($Q[0])) && GetNb ($q[0]) || 
            ($Q[1] > $q[1] || !$Q[1]) && $q[1] || ($Q[2] > $q[2] || !$Q[2]) && $q[2];
      $ERR && Exit 4;
   }
}

sub Security
{
   if ($ARGV[0] =~ /dns$/)
   {
      $ERR = 1, Error ("$MOD: \$DNSs = '$DNSs'"), 
      $ERR = 1, Error ("\$DNSs must be defined for distributed systems", 1) 
         if !$LOCAL && !$DNSs;
      $ERR && exit 1;
      my $hz = $Hz || '@';
      my $ip = $I || $HOST{$H} || @zoneCf && ("@zoneCf" =~ /^\s*(\Q$hz\E|\Q$H.\E)\s+.*\s+(a|cname)\s+([\d.]+)\s*$/im)[2];
      $ERR = 1, Error "permission denied" 
         if $UID == $RID && $ip && !AllowI $ip;
      $ERR && Exit 3;
   }
   if ($ARGV[0] !~ /dns$/)
   {
      $ERR = 1, Error ("$MOD: \$HOSTs = '$HOSTs'"), 
      $ERR = 1, Error ("\$HOSTs must include local server", 1) 
         if !$LOCAL && $HOSTs !~ /(^|\s)(\Q$HNAM\E|\Q$HIP\E)(\s|$)/ && $ARGV[0] ne "reload";
      $ERR && exit 1;
      $ERR = 1, Error "explicit ip not allowed for distributed systems or dynamic ip" 
         if $Ip && $Ip ne "." && (!$LOCAL || $IPADD == 3);
      $ERR && Exit 3;
   }
   $ERR = 1, Error ("$MOD: \$RUSER = '$RUSER'"), 
   $ERR = 1, Error ("\$RUSER can not be shared by profiles with different \$HOSTs/\$DNSs", 1) 
      if !$LOCAL && !AllowR $RUSER;
   $ERR && exit 1;

   $PERM = 4;
   $PERM = 3 if  $HOST && $HOST eq $H 
                       && (getpwnam $USER)[8] ne $SHELL[0] 
                       && (!$U || $U eq $USER);
   $PERM = 2 if !$HOST && grep (/^\s*\Q$USER\E\s+\Q$PROF\E\s/, @USER) 
                       && (getpwnam $USER)[8] ne $SHELL[0] 
                       && (!$H || !$LOH || $PROF{$H} && $OWNR{$H}) 
                       && (!$O || !$H || $O eq $OWNR{$Z}) 
                       && (!$O || $O eq $USER) 
                       && (!$I || $Ip eq ".") 
                       && (!$S || $S <= 1) 
                       && (!$A || $O eq $OWNR{GetZ $A});
   $PERM = 1 if !$HOST && grep (/^\s*\Q$USER\E\s+\Q$PROF\E\s.*\svroot\s*$/, @USER) 
                       && (getpwnam $USER)[8] ne $SHELL[0] 
                       && (!$H || !$LOH || $PROF{$H} && $OWNR{$H}) 
                       && (!$OWNR{$Z} || AllowO $OWNR{$Z}) 
                       && (!$O || !@O || AllowO $O) 
                       && (!$I || $Ip eq ".") 
                       && (!$S || $S <= $ACCESS);
   $PERM = 0 if !$HOST && ($ROOTCMD && $USER eq "root" || $ARGV[0] =~ /dns$/ && $UID == $RID) 
                       && (!$I || AllowI $I) 
                       && (!$Y || grep /^\Q$Y\E$/, keys %EVAL);
   $ERR = 1, Error "permission denied" if $PERM > $_[0];
   $ERR && Exit 3;

   Quota if $PERM;
}

my ($ok, $i);
$Ip = $I;
$I = GetI if $I eq ".";
$Z = GetZ $H;
$P = GetFn $P;
$G = GetFn $G;
$Q =~ y/kmgt/KMGT/;
$O =~ s/\./$dotChar/g, $O = substr ($O, 0, $unameLen) || $OWNR{$H};
$W =~ s/\./$dotChar/g, $W = substr $W, 0, $unameLen;
$U =~ s/\./$dotChar/g, $U = substr $U, 0, $unameLen;
@Q = split /\s+/, $Q;
@O = getpwnam $O;
@W = getpwnam $W;
@U = getpwnam $U;
@V = getpwnam $V;
$Hz = ($H =~ /^(.*)\.\Q$Z\E$/)[0];
$Iz = ($I =~ /^(\d+\.){3}(\d+)$/)[1];
$ok = $ok || /^\s*# \Q$H\E $SIGN/, $ok && /^\s*$FOOT/ && last, 
$ok && /^\s*[^#\s]+\s+(\S+)/ && ($HALL[$#HALL+1] = $1), $i++ == $#HOST && (@HALL = ()) for @HOST;
$LOH = !$H && $O || $H eq $HNAM || -c "/etc/vdata/vusers.$H";
@RUSER = getpwnam $RUSER, $RID = $RUSER[2];
@SHELL = ("!", "/bin/false", $SHELL);
@QHOST = (split /\s+/, $QHOST)[0..2];
@QUSER = (split /\s+/, $QUSER{$W || $O} || $QUSER)[0..2];
$QHARD = sprintf "%.0f", GetNb ($QUSER[0]) / $blockSize;
$QSTAB = $UNIX =~ /^Linux$/i? "usrquota" : "userquota";
$MOUNT = GetMp $HOME{$USER{$W || $O}} || $VHOME;
$UOADD = "-u $USRID -o" if $USRID;
$PAGER = $ENV{PAGER} if -f $ENV{PAGER} && -x $ENV{PAGER} && -o $ENV{PAGER};
$ZTERM = $ENV{TERM} eq "dumb" || !$ENV{TERM};
$MYCFG = $AUTOCONF || $ARGV[0] eq "autoconf" || -f "/tmp/.vhost/.ins.autoconf";
$MASIP = CheckH ($DNSs[0]) && join (".", unpack "C4", (gethostbyname $DNSs[0])[4]) || CheckI ($DNSs[0]) && $DNSs[0] || $HIP;
$DNSRE = '\s+.*\s+(ns|a|cname|ptr)\s+';
$dnsC = !$masterC || !$slaveC? $masterC || $slaveC : $MASIP eq $HIP? $masterC : $slaveC;
$zoneCf = GetZf $Z;
@zoneCf = `$RDNS cat $zoneCf $NERR` if -f $zoneCf;
$vusersCf = "/etc/vdata/vusers.$H";
$shadowCf = "/etc/vdata/shadow.$H" if !$LOH;
$passwdCf = "/etc/vdata/passwd.$H" if !$LOH;
$aliasesCf = "/etc/vdata/aliases.$H" if !$LOH;
@VUSR = Cat $vusersCf; /^\s*([^#\s]+)\s+(\S+)/ && ($VUSR{$1} ||= $2) for @VUSR;
@VSPW = Cat $shadowCf; /^\s*([^#:]+):([^:]*):/ && ($VSPW{$1} ||= $2) for @VSPW;
@VPWD = Cat $passwdCf; /^\s*([^#:]+):([^:]*:){4}([^:]*):/ && ($VPWD{$1} ||= $3) for @VPWD;
@VALI = Cat $aliasesCf; $VALI[-$_] =~ s/^(\s*,.*)//s && ($VALI[-$_-1] .= $1) for 1..$#VALI+1;
/^\s*([^#:]+):\s*(.*)/s && ($VALI{$1} ||= $2) for @VALI;
$EVAL{'$USER;'}        = 4;
$EVAL{'$PERM;'}        = 4;
$EVAL{'$HNAM;'}        = 4;
$EVAL{'$dotChar;'}     = 4;
$EVAL{'$unameLen;'}    = 4;
$EVAL{'$VUSR{$USER};'} = 3;
$EVAL{'GetU 1;'}       = 2;

Security 4         if $ARGV[0] eq "";
Security 0         if $ARGV[0] eq "addvroot";
Security 0         if $ARGV[0] eq "delvroot";
Security $HOSTADD  if $ARGV[0] eq "adddns";
Security $HOSTADD  if $ARGV[0] eq "deldns";
Security $HOSTADD  if $ARGV[0] eq "addhost";
Security $HOSTADD  if $ARGV[0] eq "delhost";
Security $HOSTADD  if $ARGV[0] eq "addhali";
Security $HOSTADD  if $ARGV[0] eq "delhali";
Security 2         if $ARGV[0] eq "adduser";
Security 2         if $ARGV[0] eq "deluser";
Security 2         if $ARGV[0] eq "adduali";
Security 2         if $ARGV[0] eq "deluali";
Security 3         if $ARGV[0] eq "passwd";
Security 2         if $ARGV[0] eq "access";
Security 1         if $ARGV[0] eq "qhost";
Security 1         if $ARGV[0] eq "quser";
Security 2         if $ARGV[0] eq "lshost";
Security 2         if $ARGV[0] eq "lsuser";
Security 2         if $ARGV[0] eq "hostlog";
Security 4         if $ARGV[0] eq "cmdlog";
Security $HOSTADD  if $ARGV[0] eq "reload";
Security 0         if $ARGV[0] eq "autoconf";
Security 0         if $ARGV[0] eq "rotatelog";
Security $EVAL{$Y} if $ARGV[0] eq "getvalue";
Security 0         if $ARGV[0] eq "uninstall";

`touch /etc/vhosts $NERR` if !-e "/etc/vhosts";
`touch /etc/vusers $NERR` if !-e "/etc/vusers";
`ln -sf /etc/vmail /etc/vdata $NERR` if !-d "/etc/vdata";
`mkdir -p /etc/vmail  $NERR` if !-e "/etc/vmail";
`mkdir -p /tmp/.vhost $NERR` if !-e "/tmp/.vhost";
`mkdir -p /var/spool/vmail $NERR` if !-e "/var/spool/vmail";
`mkdir -p /usr/local/vhost $NERR` if !-e "/usr/local/vhost";
`mkdir -p $VHOME $NERR` if !-e $VHOME;
`mkdir -p $HTEMP $NERR` if !-e $HTEMP;
`mkdir -p $UTEMP $NERR` if !-e $UTEMP;
`chown 0:$GHID /etc/vhosts $NERR` if !-o "/etc/vhosts" || (stat "/etc/vhosts")[5] != $GHID;
`chown 0:$GHID /etc/vusers $NERR` if !-o "/etc/vusers" || (stat "/etc/vusers")[5] != $GHID;
`chown 0:$GHID /etc/vdata  $NERR` if !-o "/etc/vdata"  || (lstat "/etc/vdata")[5] != $GHID;
`chown 0:$GHID /etc/vmail  $NERR` if !-o "/etc/vmail"  || (stat "/etc/vmail")[5]  != $GHID;
`chown 0:$GHID /tmp/.vhost $NERR` if !-o "/tmp/.vhost" || (stat "/tmp/.vhost")[5] != $GHID;
`chown 0:$GHID /var/spool/vmail $NERR` if !-o "/var/spool/vmail" || (stat "/var/spool/vmail")[5] != $GHID;
`chown 0:$GHID /usr/local/vhost $NERR` if !-o "/usr/local/vhost" || (stat "/usr/local/vhost")[5] != $GHID;
`chown 0:$GHID $VHOME $NERR` if !-o $VHOME || (stat $VHOME)[5] != $GHID;
`chown 0:$GHID $HTEMP $NERR` if !-o $HTEMP || (stat $HTEMP)[5] != $GHID;
`chown 0:$GHID $UTEMP $NERR` if !-o $UTEMP || (stat $UTEMP)[5] != $GHID;
`chmod 0600 /etc/vhosts $NERR` if ((stat "/etc/vhosts")[2] & 07777) != 0600;
`chmod 0600 /etc/vusers $NERR` if ((stat "/etc/vusers")[2] & 07777) != 0600;
`chmod 0710 /etc/vdata  $NERR` if ((stat "/etc/vdata")[2]  & 07777) != 0710;
`chmod 0710 /etc/vmail  $NERR` if ((stat "/etc/vmail")[2]  & 07777) != 0710;
`chmod 0710 /tmp/.vhost $NERR` if ((stat "/tmp/.vhost")[2] & 07777) != 0710;
`chmod 0711 /var/spool/vmail $NERR` if ((stat "/var/spool/vmail")[2] & 07777) != 0711;
`chmod 0755 /usr/local/vhost $NERR` if ((stat "/usr/local/vhost")[2] & 07777) != 0755;
`chmod 0711 $VHOME $NERR` if ((stat $VHOME)[2] & 07777) != 0711;
`chmod 0755 $HTEMP $NERR` if ((stat $HTEMP)[2] & 07777) != 0755;
`chmod 0755 $UTEMP $NERR` if ((stat $UTEMP)[2] & 07777) != 0755;
`ln -s /usr/bin/false /bin/false $NERR` if !-e "/bin/false";
`${FreeBSD}groupmod -g $GHID -o vhost $NERR` if getgrnam "vhost" != $GHID;
`${FreeBSD}groupadd -g $GHID -o vhost $NERR` if !getgrnam "vhost";
`${FreeBSD}groupmod -g $GUID -o vuser $NERR` if getgrnam "vuser" != $GUID;
`${FreeBSD}groupadd -g $GUID -o vuser $NERR` if !getgrnam "vuser";
`${FreeBSD}useradd -g $GHID -s $SHELL[1] nobody $NERR` if !getpwnam "nobody";
`${FreeBSD}usermod -c "vhost ruser" -d $PROF -g $GHID $RUSER $NERR` if $RUSER[3] != $GHID || $RUSER[6] ne "vhost ruser";
`${FreeBSD}useradd -c "vhost ruser" -d $PROF -g $GHID $RUSER $NERR` if !@RUSER;
$RHOME = (getpwnam $RUSER)[7];
`touch $RHOME/.rhosts $NERR` if !-e "$RHOME/.rhosts";
`chown 0:$GHID $RHOME/.rhosts $NERR` if !-o "$RHOME/.rhosts" || (stat "$RHOME/.rhosts")[5] != $GHID;
`chmod 0644 $RHOME/.rhosts $NERR` if ((stat "$RHOME/.rhosts")[2] & 07777) != 0644;
GroupAdd "nobody";

my $ok;
Flock "$cronDir/root", "+>>", 2 if !grep /^\s*$FOOT/, Cat "$cronDir/root";
BackupCf "$cronDir/root";
print FILE "$FOOT\n";
print FILE "0 5 * * 0 /usr/sbin/vhost --rotatelog -z4\n" if !grep /^[^#]*\bvhost\s+-*rotatelog\b/, @FILE;
print FILE "0 6 * * * /usr/sbin/vhost --lshost -af\n" if !grep /^[^#]*\bvhost\s+-*lshost\b/, @FILE;
print FILE "5 6 * * * /usr/sbin/vhost --lsuser\n" if !grep /^[^#]*\bvhost\s+-*lsuser\b/, @FILE;
`touch $cronDir $NERR` if -f FILE;
Flock 0;
Flock "/etc/vhosts", ">>", 2 if !-s "/etc/vhosts";
print FILE "### /etc/vhosts - vhost host index, do not edit\n";
print FILE "#\n";
print FILE "# ip_address\thost_name mailbox_size #user #uali\n";
print FILE "\n";
print FILE "127.0.0.1\tlocalhost.localdomain localhost\n";
Flock 0;
Flock "/etc/vusers", ">>", 2 if !-s "/etc/vusers";
print FILE "### /etc/vusers - vhost user index, do not edit\n";
print FILE "#\n";
print FILE "# user_name\tprofile_dir disk_space #host #hali vroot\n";
print FILE "\n";
Flock 0;
Flock "/etc/shells", ">>", 2 if !grep /^\s*\Q$SHELL[1]\E\s*$/, Cat "/etc/shells";
BackupCf "/etc/shells";
print FILE "$SHELL[1]\n";
Flock 0;
Flock "/etc/fstab", "+<", 0 if !grep /^\s*[^#\s]+\s+\Q$MOUNT\E\s.*\b\Q$QSTAB\E\b/, Cat "/etc/fstab";
BackupCf "/etc/fstab";
s/^(\s*[^#\s]+\s+\Q$MOUNT\E\s+\S+\s+\S+)/$1,$QSTAB/ && ($ok = 1) for @FILE;
print FILE @FILE;
Flock 0;
Flock "/etc/mtab", "+<", 0 if $ok && -f "/etc/mtab" && !grep /^\s*[^#\s]+\s+\Q$MOUNT\E\s.*\b\Q$QSTAB\E\b/, Cat "/etc/mtab";
s/^(\s*[^#\s]+\s+\Q$MOUNT\E\s+\S+\s+\S+)/$1,$QSTAB/ for @FILE;
print FILE @FILE;
Flock 0;
Flock "$VHOME/.login", ">>", 2 if !-s "$VHOME/.login";
print FILE "set prompt=\"`uname -n`> \"\nset path=(\$path /sbin /usr/sbin)\n";
Flock 0;
Flock "$VHOME/.profile", ">>", 2 if !-s "$VHOME/.profile";
print FILE "PS1=\"`uname -n`> \"\nPATH=\$PATH:/sbin:/usr/sbin\n";
Flock 0;
Flock "$RHOME/.rhosts", ">", 0 if Cat ("$RHOME/.rhosts") ne 
           join "\n", @HOSTs, $DNSs =~ /(^|\s)(\Q$HNAM\E|\Q$HIP\E)(\s|$)/ && @DNSs, "";
print FILE "$_\n" for @HOSTs, $DNSs =~ /(^|\s)(\Q$HNAM\E|\Q$HIP\E)(\s|$)/ && @DNSs;
Flock 0;
`quotaoff   $MOUNT $NERR` if $ok;
`quotacheck $MOUNT $NERR` if $ok;
`quotaon    $MOUNT $NERR` if $ok;

if ($H && !$HLP && $ARGV[0] !~ /^(.*dns|addhost|delhost|delhali)$/)
{
   my $loh = $LOH if $ARGV[0] !~ /^(addhali|qhost|lshost|hostlog)$/;
   Error5 "'$H' is an alias" if $HOST{$H} && !$QHOST{$H};
   Error5 "'$H' does not exist" if !$loh && !$QHOST{$H};
   Error5 "'$H' is a deleted host" if !$loh && !CheckQ $QHOST{$H};
}
if ($O && !$H && $ARGV[0] =~ /^(quser|lshost|lsuser)$/)
{
   Error5 "'$O' is not a host owner" if !@O || !$USER{$O};
}
if ($ARGV[0] =~ /^(deluser|passwd|access|quser)$/)
{
   my $ruser = "vhost ruser" if $ARGV[0] ne "passwd";
   Error5 "'$U".($H&&'@')."$H' is an alias" if $VALI{$U} && !($LOH && @U) && !$VPWD{$U};
   Error5 "'$U".($H&&'@')."$H' does not exist" if $U && !($LOH && @U) && !$VPWD{$U};
   Error5 "permission denied" if $VUSR{$U} eq "root" || 
      $LOH && ($U eq "root" || $U[3] == $GUID && $U[6] eq "vhost vuser" || $U[3] == $GHID && $U[6] eq $ruser) || 
      $ARGV[0] ne "deluser" && $U eq $O && $O ne $USER && $USER ne "root" && grep (/^\s*\Q$U\E\s.*\svroot\s*$/, @USER) || 
      $ARGV[0] eq "quser" && $O ne $USER && $USER ne "root" && grep /^\s*\Q$O\E\s.*\svroot\s*$/, @USER;
}
if ($ARGV[0] =~ /^(adduali|deluali)$/)
{
   Error5 "'$H' is not virtual" if $LOH && !-c "/etc/vdata/vusers.$H" && $V eq "catch-all@";
}


};
$X eq "addvroot" && do { #####################################################

Error5 "'$P' already exists" if !$W && -e $P;
Error5 "'$P' is not a profile" if $W && -e $P && !-f "$P/V00hostconf";
Error5 "'$W' is a system user" if $W eq "root" || $W[3] == $GUID && $W[6] eq "vhost vuser";
Error5 "'$W' at '$P' already exists" if $W && grep /^\s*\Q$W\E\s+\Q$P\E\s.*\svroot\s*$/, @USER;
Prompt "creating '$P'" if !-e $P;
Prompt "'$W' already exists" if $W && @W;
my $pass = Passwd "password for '$W'" if $W && !@W;
$USRID = GetU if !@W && !$LOCAL && $UIDSYNC;
Error5 "'$USRID' already exists as a uid" if $USRID && ($USRID < $UIDSYNC || getpwuid $USRID);
$LOCAL = 1, Rexec "vhost --addvroot $P $W $_f", "HOST", $HOSTs if !$LOCAL;

my $ok;
my $p = GetLn $PROF;
`ln -s $p $P $NERR` if !-e $P && !$_c;
`cp -pR $p/. $P $NERR` if !-e $P && $_c;
`${FreeBSD}usermod -d $1 $W $NERR` if $W[3] == $GUID && $W[7] =~ /(.+)\/[^\/]+\/?$/ && grep /^\Q$1\E$/, values %HOME;
`${FreeBSD}useradd -c "vhost admin" -d $VHOME -g $GUID -s $SHELL[$ACCESS] $UOADD $W $NERR`;
Quota::setqlim (Quota::getqcarg ($MOUNT), (getpwnam $W)[2], 0, $QHARD, 0, 0) if $W && !@W;
SetPass $W, $pass if $W && !@W;
@W = getpwnam $W;
$QUSER[0] = GetKb GetNb Quota::query (Quota::getqcarg ($MOUNT), $W[2]) && $QUSER[0] || "0K";
Flock "/etc/vusers", "+<", 0;
$W && s/^(\s*\Q$W\E\s+\Q$P\E\s.*\d).*/$1 vroot/ && ($ok = 1) for @FILE;
$W && ($FILE[$#FILE+1] = sprintf "%-15s $P @QUSER vroot\n", $W) if !$ok;
print FILE @FILE;
Flock 0;


};
$X eq "delvroot" && do { #####################################################

Error5 "'$P' does not exist" if !-e $P;
Error5 "'$P' is not a profile" if -e $P && !-f "$P/V00hostconf";
Error5 "'$P' is the default profile" if !$W && $P eq "/etc/vhost.d";
Error5 "'$P' has associated hosts" if !$W && grep $PROF{$_} eq $P, keys %PROF;
Error5 "'$W' at '$P' does not exist" if $W && !grep /^\s*\Q$W\E\s+\Q$P\E\s.*\svroot\s*$/, @USER;
Prompt "deleting all at '$P'" if !$W;
Prompt "deleting '$W' at '$P'" if $W;
$LOCAL = 1, Rexec "vhost --delvroot $P $W $_f", "HOST", $HOSTs if !$LOCAL;

my $w;
`rm -r $P $NERR` if !$W;
Flock "/etc/vusers", "+<", 0;
$W || s/^\s*([^#\s]+)\s+\Q$P\E\s.*//s && ($w .= "$1 ") for @FILE;
$W && s/^(\s*\Q$W\E\s+\Q$P\E\s.*)\svroot.*/$1/ for @FILE;
print FILE @FILE;
Flock 0;
CleanO $_ for $W || split /\s+/, $w;


};
$X eq "adddns" && do { #######################################################

Error5 "dns zone '$Z' does not exist" if $Hz && !@zoneCf;
Prompt "dns zone '$Z' already exists" if !$Hz && @zoneCf;
Prompt "dns record '$H' already exists" if $Hz && grep /^\s*(\Q$Hz\E|\Q$H.\E)$DNSRE/i, @zoneCf;
Prompt "reverse mapping already exists" if $_i && grep /^\s*(\Q$Iz\E|\Q$I.\E)$DNSRE/i, @revZf;
$CHILD = 1, Rexec "vhost --adddns $H $_z $_i $_q $_f", "HOST", $HOSTs 
   if !$LOCAL && !$CHILD && $Ip eq "." && $HOSTs =~ /(^|\s)(\Q$HNAM\E|\Q$HIP\E)(\s|$)/;
$LOCAL = 1, Rexec "vhost --adddns $H $I $_z $_i $_q $_f", "DNS", $DNSs if !$LOCAL;

if (!$Hz)
{
   AddConf $dnsCf, $dnsC;
   @dnsCf = Cat $dnsCf;
   $zoneCf = GetZf $Z;
   my $fs = UpdateFs $zoneCf;
   $zoneC =~ s/\[NS_NAME\]/$HNAM/g;
   $zoneC =~ s/\[ZONE_NAME\]/$Z/g;
   $zoneC =~ s/\[IP_ADDRESS\]/$I/g;
   $zoneC =~ s/\[FILE_SERIAL\]/$fs/g;
   `touch $zoneCf $NERR`;
   Flock $zoneCf, "+<", 0;
   UpdateFs;
   print FILE @FILE;
   @FILE || print FILE $zoneC;
   @FILE && "@FILE" !~ /^\s*\Q$_\E/im && print FILE for grep /\s(a|cname)\s/i, split /^/, $zoneC;
   Flock 0;
}
if ($Hz)
{
   Flock $zoneCf, "+<", 0;
   UpdateFs;
   print FILE @FILE;
   print FILE "$Hz\t\tIN\tA\t$I\n" if !grep /^\s*\Q$Hz\E\s+.*\s+(a|cname)\s+\Q$I\E\s*$/i, @FILE;
   Flock 0;
}
if ($_i)
{
   Flock $revZf, "+<", 0;
   UpdateFs;
   print FILE @FILE;
   print FILE "$Iz\t\tIN\tPTR\t$H.\n" if !grep /^\s*\Q$Iz\E\s+.*\s+(ptr)\s+\Q$H.\E\s*$/i, @FILE;
   Flock 0;
}
ReloadRc 0;
ReloadRc $dnsRc if !$_q;


};
$X eq "deldns" && do { #######################################################

Error5 "dns zone '$Z' does not exist" if !@zoneCf;
if (!$Hz && !$I && grep /^\s*([^\s\@]+)$DNSRE/i && $zoneC !~ /^\s*\Q$1\E$DNSRE/im, @zoneCf) {
   Error5 "dns zone '$Z' contains host records";
} elsif (!$Hz && !$I) {
   Prompt "deleting dns zone '$Z'";
}
if (!$Hz && $I && !grep /^\s*(\@|\Q$Z.\E)$DNSRE\Q$I\E\s*$/i, @zoneCf) {
   Error5 "dns zone '$Z->$I' does not exist";
} elsif (!$Hz && $I) {
   Prompt "deleting dns zone '$Z->$I'";
}
if ($Hz && !$I && !grep /^\s*(\Q$Hz\E|\Q$H.\E)$DNSRE/i, @zoneCf) {
   Error5 "dns record '$H' does not exist";
} elsif ($Hz && !$I) {
   Prompt "deleting dns record '$H'";
}
if ($Hz && $I && !grep /^\s*(\Q$Hz\E|\Q$H.\E)$DNSRE\Q$I\E\s*$/i, @zoneCf) {
   Error5 "dns record '$H->$I' does not exist";
} elsif ($Hz && $I) {
   Prompt "deleting dns record '$H->$I'";
}
$LOCAL = 1, Rexec "vhost --deldns $H $I $_q $_f", "DNS", $DNSs if !$LOCAL;

if (!$Hz)
{
   Flock $zoneCf, "+<", 0;
   UpdateFs;
   $I && /^\s*(\S+)$DNSRE\Q$I\E\s*$/i && $zoneC =~ /^\s*\Q$1\E$DNSRE/im && s/.*//s for @FILE;
   print FILE @FILE;
   Flock 0;
}
if ($Hz)
{
   Flock $zoneCf, "+<", 0;
   UpdateFs;
   $I || s/^\s*(\Q$Hz\E|\Q$H.\E)$DNSRE.*//is for @FILE;
   $I && s/^\s*(\Q$Hz\E|\Q$H.\E)$DNSRE\Q$I\E\s*$//is for @FILE;
   print FILE @FILE;
   Flock 0;
}
if (@revZf)
{
   Flock $revZf, "+<", 0;
   UpdateFs;
   $I || s/^\s*[\d.]+$DNSRE\Q$H.\E\s*$//is for @FILE;
   $I && s/^\s*(\Q$Iz\E|\Q$I.\E)$DNSRE\Q$H.\E\s*$//is for @FILE;
   print FILE @FILE;
   Flock 0;
}
if (!$Hz && !$I || !grep /^\s*[^\s\@]+$DNSRE/i, Cat $zoneCf)
{
   DelConf $dnsCf, $dnsC;
   my $zf = (split /\//, $zoneCf)[-1];
   `mv $zoneCf /tmp/.vhost/.$zf $NERR`;
}
ReloadRc 0;
ReloadRc $dnsRc if !$_q;


};
$X eq "addhost" && do { ######################################################

Error5 "'$H' already exists as an alias" if $HOST{$H} && !$QHOST{$H};
Error5 "'$H' already exists" if CheckQ ($QHOST{$H}) && (!$_f || $O ne $OWNR{$H});
Error5 "'$O' is a system user" if !$LOH && $O eq "root" || $O[3] == $GUID && $O[6] eq "vhost vuser";
Error5 "'$O' can not occupy multiple \$VHOME's" if $HOME{$USER{$O}} && $HOME{$USER{$O}} ne $VHOME;
Prompt "'$O' already exists" if @O && $O ne $USER;
my $pass = Passwd "password for '$O'" if !@O;
$USRID = GetU if !@O && !$LOCAL && $UIDSYNC;
Error5 "'$USRID' already exists as a uid" if $USRID && ($USRID < $UIDSYNC || getpwuid $USRID);
$LOCAL = 1, Rexec "vhost --addhost $H $O $_q $_f", "HOST", $HOSTs if !$LOCAL;
$DNSAWARE && $HOST{$H} && $HOST{$H} ne $I && Rdns "vhost --deldns $_ $HOST{$H} $_q -f" for @HALL;
$DNSAWARE && Rdns "vhost --adddns $_ $I $_z $_i $_q -f" for !@HALL && $H || @HALL;

my $ok;
`mv $HOME{$PROF{$H}}/$H $VHOME $NERR`;
`mkdir $VHOME/$H $NERR`;
`mkdir $VHOME/$H/home $NERR`;
`mkdir $VHOME/$H/bin $NERR`;
`mkdir $VHOME/$H/etc $NERR`;
`mkdir $VHOME/$H/lib $NERR`;
`mkdir $VHOME/$H/var $NERR`;
`mkdir $VHOME/$H/var/log $NERR`;
`mkdir $VHOME/$H/var/mail $NERR`;
`touch $VHOME/$H/var/log/.no-rotate $NERR`;
`chmod 0750 $VHOME/$H $NERR`;
`chmod 1777 $VHOME/$H/var/mail $NERR`;
`rm -r /etc/vdata/vhostd.$H $NERR`;
`ln -s $PROF /etc/vdata/vhostd.$H $NERR`;
`touch /etc/vdata/vusers.$H $NERR`;
`touch /etc/vdata/shadow.$H $NERR`;
`touch /etc/vdata/passwd.$H $NERR`;
`touch /etc/vdata/aliases.$H $NERR`;
`${FreeBSD}usermod -d $VHOME $O $NERR` if $O[3] == $GUID && $O ne $OWNR{$H};
`${FreeBSD}useradd -c "vhost admin" -d $VHOME/$H -g $GUID -s $SHELL[$ACCESS] $UOADD $O $NERR`;
Quota::setqlim (Quota::getqcarg ($MOUNT), (getpwnam $O)[2], 0, $QHARD, 0, 0) if !@O;
SetPass $O, $pass if !@O;
@O = getpwnam $O;
$QUSER[0] = GetKb GetNb Quota::query (Quota::getqcarg ($MOUNT), $O[2]) && $QUSER[0] || "0K";
$QHOST[0] = GetKb GetNb $QHOST[0];
AddConf "/etc/vhosts", "$I\t$H @QHOST\n";
Flock "/etc/vhosts", "+<", 0;
$ok = $ok || /^\s*# \Q$H\E $SIGN/, $ok && /^\s*$FOOT/ && last, $ok && s/^(\s*[^#\s]+\s+\Q$H\E)\s+-d.*/$1 @QHOST\n/s for @FILE;
print FILE @FILE;
Flock 0;
Flock "/etc/vusers", "+>>", 2;
printf FILE "%-15s $PROF @QUSER\n", $O if $O ne "root" && !grep /^\s*\Q$O\E\s+\Q$PROF\E\s/, @FILE;
Flock 0;
Flock "$VHOME/$H/.login", "+>>", 2;
print FILE "set pwd=`pwd`; set prompt=\"`basename \$pwd`> \"\nset path=(\$path /sbin /usr/sbin)\n" if !@FILE;
Flock 0;
Flock "$VHOME/$H/.profile", "+>>", 2;
print FILE "pwd=`pwd`; PS1=\"`basename \$pwd`> \"\nPATH=\$PATH:/sbin:/usr/sbin\n" if !@FILE;
Flock 0;
CleanO $OWNR{$H} if $O ne $OWNR{$H};

if ($LOH)
{
   `chmod 1777 $mailDir $NERR`;
   `rm -r /var/spool/vmail/$H $NERR`;
   `ln -s $mailDir /var/spool/vmail/$H $NERR`;
   `ln -s $homeDir/$O $VHOME/$H/home/$O $NERR` if $O[7] eq "$homeDir/$O" && !-e "$VHOME/$H/home/$O";
   `mkdir $VHOME/$H/home/$O $NERR`;
   `ln -sf $mailDir/$O $VHOME/$H/var/mail/$O $NERR`;
   `ln -sf /dev/null /etc/vdata/vusers.$H $NERR`;
   `ln -sf $shadowCf /etc/vdata/shadow.$H $NERR`;
   `ln -sf $passwdCf /etc/vdata/passwd.$H $NERR`;
   `ln -sf $aliasesCf /etc/vdata/aliases.$H $NERR`;
} else {
   $_ ne $O && $_ ne $OWNR{$H} || next, 
   `ln -sf $VHOME/$H/var/mail/$_ $mailDir/$VUSR{$_} $NERR`, 
   `${FreeBSD}useradd -c "vhost vuser" -d $VHOME/$H/home/$_ -g $GUID -s $SHELL[1] -u $O[2] -o $VUSR{$_} $NERR`, 
      for keys %VUSR;
   `rm -r /var/spool/vmail/$H $NERR`;
   `ln -s $VHOME/$H/var/mail /var/spool/vmail/$H $NERR`;
   `mkdir $VHOME/$H/home/$O $NERR`;
   `ln -sf $mailDir/$O $VHOME/$H/var/mail/$O $NERR`;
   my $days = int time / 86400;
   my $pswd = $pass && Crypt $pass;
   $pswd = $pswd || $O[1] || "*", $pswd = $O[8] eq $SHELL[0] && "$SHELL[0]$pswd" || $pswd;
   Flock "/etc/vdata/vusers.$H", "+<", 0 if $O ne $OWNR{$H};
   s/^\s*\Q$O\E\s.*//s for @FILE;
   $FILE[0] = "";
   printf FILE "%-15s %s\n", $O, $O;
   print FILE @FILE;
   Flock 0;
   Flock "/etc/vdata/shadow.$H", "+<", 0 if $O ne $OWNR{$H};
   s/^\s*\Q$O\E:.*//s for @FILE;
   $FILE[0] = "";
   print FILE "$O:$pswd:$days:-1:99999:-1:-1:-1:\n";
   print FILE @FILE;
   Flock 0;
   Flock "/etc/vdata/passwd.$H", "+<", 0 if $O ne $OWNR{$H};
   s/^(\s*([^:]*:)[^:]*:)[^:]*:([^:]*:[^:]*:)[^:]*:(.*)/$1$O[2]:$3$VHOME\/$H\/home\/$2$4/, 
   s/^\s*\Q$O\E:.*//s for @FILE;
   $FILE[0] = "";
   print FILE "$O:$pswd:$O[2]:$GUID:vhost admin:$VHOME/$H:$SHELL[1]\n";
   print FILE @FILE;
   Flock 0;
   Flock "/etc/vdata/aliases.$H", "+<", 0 if $O ne $OWNR{$H};
   s/^\s*\Q$ADMIN\E:.*//s for @FILE;
   print FILE "$ADMIN:  $O \n";
   print FILE @FILE;
   Flock 0;
}


};
$X eq "delhost" && do { ######################################################

Error5 "'$H' is an alias" if $HOST{$H} && !$QHOST{$H};
Error5 "'$H' does not exist" if !$QHOST{$H};
Prompt "deleting '$H'";
Forced "delete all host data";
$LOCAL = 1, Rexec "vhost --delhost $H $_q $_f $_d", "HOST", $HOSTs if !$LOCAL;
$DNSAWARE && $_d && Rdns "vhost --deldns $_ $HOST{$H} $_q -f" for @HALL;

my $ok;
`chmod 0000 $HOME{$PROF{$H}}/$H $NERR`;
Flock "/etc/vhosts", "+<", 0;
$ok = $ok || /^\s*# \Q$H\E $SIGN/, $ok && /^\s*$FOOT/ && last, $ok && s/^(\s*[^#\s]+\s+\Q$H\E)\s.*/$1 -d\n/s for @FILE;
print FILE @FILE;
Flock 0;
$_ ne $O && $VUSR[0] !~ /^\s*\Q$_\E\s/ || next, 
`rm $mailDir/$VUSR{$_} $NERR`, 
`${FreeBSD}userdel $VUSR{$_} $NERR` 
   for keys %VUSR;
if ($_d)
{
   `rm -r /var/spool/vmail/$H $NERR`;
   `rm -r $HOME{$PROF{$H}}/$H $NERR` if $HTEMP ne "$HOME{$PROF{$H}}/$H";
   `rm -r /etc/vdata/vhostd.$H $NERR`;
   `rm /etc/vdata/vusers.$H $NERR`;
   `rm /etc/vdata/shadow.$H $NERR`;
   `rm /etc/vdata/passwd.$H $NERR`;
   `rm /etc/vdata/aliases.$H $NERR`;
   DelConf "/etc/vhosts", "$HOST{$H}\t$H @QHOST\n";
   CleanO $O;
}


};
$X eq "addhali" && do { ######################################################

Error5 "'$A' already exists" if $HOST{$A} && !$QHOST{$A};
Error5 "'$A' already exists as a host" if $QHOST{$A};
$LOCAL = 1, Rexec "vhost --addhali $H $A $_q", "HOST", $HOSTs if !$LOCAL;
$DNSAWARE && Rdns "vhost --adddns $_ $HOST{$H} $_z $_q -f" for $A;

my $ok;
Flock "/etc/vhosts", "+<", 0;
$ok = $ok || /^\s*# \Q$H\E $SIGN/, $ok && s/^(\s*$FOOT)/$HOST{$H}\t$A\n$1/ && last for @FILE;
print FILE @FILE;
Flock 0;


};
$X eq "delhali" && do { ######################################################

Error5 "'$A' does not exist" if !$HOST{$A};
Error5 "'$A' is a host" if $QHOST{$A};
$LOCAL = 1, Rexec "vhost --delhali $A $_q", "HOST", $HOSTs if !$LOCAL;
$DNSAWARE && Rdns "vhost --deldns $_ $HOST{$H} $_q -f" for $A;

my $ok;
Flock "/etc/vhosts", "+<", 0;
$ok = $ok || /^\s*# \Q$H\E $SIGN/, $ok && /^\s*$FOOT/ && last, $ok && s/^\s*[^#\s]+\s+\Q$A\E\s*$//s for @FILE;
print FILE @FILE;
Flock 0;


};
$X eq "adduser" && do { ######################################################

Error5 "'$U".($H&&'@')."$H' already exists" if ($LOH && @U || $VPWD{$U}) && !$_f;
Print "warning - '$U".($H&&'@')."$H' already exists as an alias" if $VALI{$U};
my $pass = Passwd "password for '$U".($H&&'@')."$H'";
if (!$LOH)
{
   my $vusr = substr "$U.$H", 0, $unameLen; $vusr =~ s/\./$dotChar/g;
   print "vhost: alternative login [$vusr]: ";
   chomp ($VUSR = substr <STDIN>, 0, $unameLen), $VUSR =~ s/\./$dotChar/g if !$_f;
   print "$VUSR\n" if !$_f && $ZTERM;
   print "\n" if $_f;
   $RECHO .= "$VUSR\n" if !$_f;
   $VUSR = $VUSR || $VUSR{$U} || $vusr;
   $ERR = !CheckU $VUSR, 1;
   $ERR && Exit 5;
   Error5 "alternative login too short ($minLogin characters minimum)" if length $VUSR < $minLogin;
   Error5 "'$VUSR' already exists" if (getpwnam $VUSR)[0] && !$VPWD{$U};
}
my $n = $N;
s/([\\\$"`])/\\\\\\$1/g, s/(.+)/"\\"$1\\""/ for $n;
$LOCAL = 1, Rexec "vhost --adduser $H $U $n $_f", "HOST", $HOSTs if !$LOCAL;

my $n = $N;
s/:/;/g for $N;
s/:/;/g, s/([\\\$"`])/\\$1/g for $n;
if ($LOH)
{
   `mkdir $homeDir/$U $NERR`;
   `chmod 0755 $homeDir/$U $NERR`;
   `rm -r $VHOME/$H/home/$U $NERR`;
   `ln -s $homeDir/$U $VHOME/$H/home/$U $NERR`;
   `ln -sf $mailDir/$U $VHOME/$H/var/mail/$U $NERR`;
#  `touch $mailDir/$U $NERR`;
   `${FreeBSD}useradd -c "$n" -d $homeDir/$U -s $SHELL[$ACCESS] $U $NERR`;
   SetPass $U, $pass;
   @U = getpwnam $U;
} else {
   `mkdir $VHOME/$H/home/$U $NERR`;
   `chmod 0755 $VHOME/$H/home/$U $NERR`;
   `ln -sf $VHOME/$H/var/mail/$U $mailDir/$VUSR $NERR`;
   `touch $VHOME/$H/var/mail/$U $NERR`;
   `${FreeBSD}useradd -c "vhost vuser" -d $VHOME/$H/home/$U -g $GUID -s $SHELL[1] -u $O[2] -o $VUSR $NERR`;
   SetPass $VUSR, $pass;
   my $days = int time / 86400;
   my $pswd = $pass && Crypt $pass;
   $pswd = $pswd || "*";
   Flock "/etc/vdata/vusers.$H", "+>>", 2;
   printf FILE "%-15s %s\n", $U, $VUSR if !grep /^\s*\Q$U\E\s/, @FILE;
   Flock 0;
   Flock "/etc/vdata/shadow.$H", "+>>", 2;
   print FILE "$U:$pswd:$days:-1:99999:-1:-1:-1:\n" if !grep /^\s*\Q$U\E:/, @FILE;
   Flock 0;
   Flock "/etc/vdata/passwd.$H", "+>>", 2;
   print FILE "$U:$pswd:$O[2]:$GUID:$N:$VHOME/$H/home/$U:$SHELL[1]\n" if !grep /^\s*\Q$U\E:/, @FILE;
   Flock 0;
}


};
$X eq "deluser" && do { ######################################################

Error5 "'$U' is the host owner" if $U eq $O;
Prompt "deleting '$U".($H&&'@')."$H'";
Forced "delete all user data";
$LOCAL = 1, Rexec "vhost --deluser $H $U $_f $_d", "HOST", $HOSTs if !$LOCAL;

if ($LOH)
{
   `chmod 0000 $homeDir/$U $NERR`;
   `rm -r $homeDir/$U $NERR` if $_d && $UTEMP ne "$homeDir/$U";
   `rm $mailDir/$U $NERR` if $_d;
   `rm $VHOME/$H/home/$U $NERR`;
   `rm $VHOME/$H/var/mail/$U $NERR`;
   `${FreeBSD}userdel $U $NERR`;
} else {
   `chmod 0000 $VHOME/$H/home/$U $NERR`;
   `rm -r $VHOME/$H/home/$U $NERR` if $_d && $UTEMP ne "$HTEMP/home/$U";
   `rm $VHOME/$H/var/mail/$U $NERR` if $_d;
   `rm $mailDir/$VUSR{$U} $NERR`;
   `${FreeBSD}userdel $VUSR{$U} $NERR`;
   Flock "/etc/vdata/vusers.$H", "+<", 0;
   s/^\s*\Q$U\E\s.*//s for @FILE;
   print FILE @FILE;
   Flock 0;
   Flock "/etc/vdata/shadow.$H", "+<", 0;
   s/^\s*\Q$U\E:.*//s for @FILE;
   print FILE @FILE;
   Flock 0;
   Flock "/etc/vdata/passwd.$H", "+<", 0;
   s/^\s*\Q$U\E:.*//s for @FILE;
   print FILE @FILE;
   Flock 0;
}


};
$X eq "adduali" && do { ######################################################

Print "warning - '$V".($H&&'@')."$H' already exists as a user" if ($LOH && @V || $VPWD{$V}) && !$VALI{$V};
my @n = @N;
s/([\\\$"`])/\\\\\\$1/g, s/(.+)/"\\"$1\\""/ for @n;
$LOCAL = 1, Rexec "vhost --adduali $H $V @n", "HOST", $HOSTs if !$LOCAL;

s/,/./g, s/\s+/ /g, s/^\s*(\|.*)/"$1"/ for @N;
my $ok;
my $vali = $VALI{$V} || "$N[0] ";
$vali =~ s/\s+/ /g;
$vali !~ /(^|,)\s*\Q$_\E\s*(,|$)/ && ($vali .= ",$_ ") for @N;
$vali =~ s/^[\s,]*//;
$vali = $N[0] if $V eq "catch-all@";
Flock $aliasesCf, "+<", 0;
$FILE[-$_] =~ s/^(\s*,.*)//s && ($FILE[-$_-1] .= $1), 
$FILE[-$_] =~ s/^(\s*\Q$V\E:).*/$1  $vali\n/s && ($ok = 1) for 1..$#FILE+1;
$FILE[$#FILE+1] = "$V:  $vali\n" if !$ok;
print FILE @FILE;
Flock 0;


};
$X eq "deluali" && do { ######################################################

my @n = @N;
s/([\\\$"`])/\\\\\\$1/g, s/(.+)/"\\"$1\\""/ for @n;
$LOCAL = 1, Rexec "vhost --deluali $H $V @n", "HOST", $HOSTs if !$LOCAL;

s/,/./g, s/\s+/ /g, s/^\s*(\|.*)/"$1"/ for @N;
my $vali = $VALI{$V} if @N;
$vali =~ s/\s+/ /g;
$vali =~ s/(^|,)\s*\Q$_\E\s*(,|$)/$2/ for @N;
$vali =~ s/^[\s,]*//;
Flock $aliasesCf, "+<", 0;
$FILE[-$_] =~ s/^(\s*,.*)//s && ($FILE[-$_-1] .= $1), 
$FILE[-$_] =~ s/^(\s*\Q$V\E:).*/$1  $vali\n/s && !$vali && ($FILE[-$_] = "") for 1..$#FILE+1;
print FILE @FILE;
Flock 0;


};
$X eq "passwd" && do { #######################################################

my $pass = $N || Passwd "password for '$U".($H&&'@')."$H'";
my $n = $N;
s/([\\\$"`])/\\\\\\$1/g, s/(.+)/"\\"$1\\""/ for $n;
$LOCAL = 1, Rexec "vhost --passwd $H $U $n", "HOST", $HOSTs if !$LOCAL;

my $pswd = $pass && Crypt $pass;
for my $u ($pswd && $U) {
for ($u ne $O && !$LOH && $u && $H || grep $OWNR{$_} eq $u, keys %PROF)
{
   Flock "/etc/vdata/shadow.$_", "+<", 0;
   s/^(\s*\Q$u\E:!*(\Q$SHELL[0]\E)*)[^:]*(:.*)/$1$pswd$3/ for @FILE;
   print FILE @FILE;
   Flock 0;
   Flock "/etc/vdata/passwd.$_", "+<", 0;
   s/^(\s*\Q$u\E:!*(\Q$SHELL[0]\E)*)[^:]*(:.*)/$1$pswd$3/ for @FILE;
   print FILE @FILE;
   Flock 0;
}
my $v = !$LOH? $VUSR{$u} : $u;
SetPass $v, $pass;
}


};
$X eq "access" && do { #######################################################

Error5 "'$U' is yourself" if $U eq $O && $O eq $USER;
$LOCAL = 1, Rexec "vhost --access $H $U $S", "HOST", $HOSTs if !$LOCAL;

my $pswd = $SHELL[0] if !$S;
for my $u ($U || !$LOH && grep $_ ne $O, keys %VPWD) {
for ($u ne $O && !$LOH && $u && $H || grep $OWNR{$_} eq $u, keys %PROF)
{
   Flock "/etc/vdata/shadow.$_", "+<", 0;
   s/^(\s*\Q$u\E:)!*(\Q$SHELL[0]\E)*(.*)/$1$pswd$3/ for @FILE;
   print FILE @FILE;
   Flock 0;
   Flock "/etc/vdata/passwd.$_", "+<", 0;
   s/^(\s*\Q$u\E:)!*(\Q$SHELL[0]\E)*(.*)/$1$pswd$3/ for @FILE;
   print FILE @FILE;
   Flock 0;
}
my $s = $u eq $O || $LOH? $S : !$S? $S : 1;
my $v = !$LOH? $VUSR{$u} : $u;
`${FreeBSD}usermod -s $SHELL[$s] $v $NERR`;
}


};
$X eq "qhost" && do { ########################################################

$LOCAL = 1, Rexec "vhost --qhost $H $Q", "HOST", $HOSTs if !$LOCAL;

$Q[0] = GetKb GetNb $Q[0];
Flock "/etc/vhosts", "+<", 0;
s/^(\s*[^#\s]+\s+\Q$H\E)\s.*/$1 @Q\n/s for @FILE;
print FILE @FILE;
Flock 0;


};
$X eq "quser" && do { ########################################################

my @qu = Quota::query (Quota::getqcarg ($MOUNT), $O[2]);
Error5 "disk quota not enabled on file system '$MOUNT'" if GetNb ($Q[0]) && !@qu;
$LOCAL = 1, Rexec "vhost --quser $O $Q", "HOST", $HOSTs if !$LOCAL;

$Q[0] = GetKb GetNb $Q[0];
Quota::setqlim (Quota::getqcarg ($MOUNT), $O[2], 0, sprintf ("%.0f", GetNb ($Q[0]) / $blockSize), $qu[5], $qu[6]);
Flock "/etc/vusers", "+<", 0;
s/^(\s*\Q$O\E\s+\S+)\s.*\d(.*)/$1 @Q$2/ for @FILE;
print FILE @FILE;
Flock 0;


};
$X eq "lshost" && do { #######################################################

my ($h, $hl, $al, $hsort, $asort, $hnum, $anum, $hcmp, $acmp, $hmya, $amya, $hmyb, $amyb, @q, @hl, @al, @date, %du, %dv);
`rm /tmp/.vhost/.duh $NERR` if !$H && $USER eq "root" && $_f;
/^\s*([^#\s]+)\s+(\S+)\s+(\S+)/ && ($du{$3} = $1, $dv{$3} = $2) for $DUCACHE && Cat "/tmp/.vhost/.duh";
$hnum = (9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12)[$_n<13?$_n:13];
$anum = (0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 1, 10, 11, 12)[$_n<13?$_n:13];
$hmya = "(split /[\\/\\s]+/, \$a, 12)[$hnum]", $hmyb = "(split /[\\/\\s]+/, \$b, 12)[$hnum]";
$amya = "(split /[\\/\\s]+/, \$a,  2)[$anum]", $amyb = "(split /[\\/\\s]+/, \$b,  2)[$anum]";
$hmya = "GetNb ($hmya)", $hmyb = "GetNb ($hmyb)" if $_n =~ /^[34]$/;
$hcmp = $_n =~ /^[345678]$/? "<=>" : "cmp";
$acmp = "cmp";
$hsort = ($_r && "reverse ") . ($_n != 9 && "sort {$hmya $hcmp $hmyb} ") if "$_s$_r$_n";
$asort = ($_r && "reverse ") . ($_n != 9 && "sort {$amya $acmp $amyb} ") if "$_s$_r$_n" && $anum < 2;

for (@HOST)
{
   /^\s*$HEAD|^\s*$FOOT|^\s*# (\S+) $SIGN\s+(.*)|^\s*[^#\s]+\s+(\S+)/ && !$3 && ($h = $1, @date = split /\s+/, $2);
   next if !$h || !$3 || $H && $h ne $H || !$_a && !CheckQ ($QHOST{$h}) || 
      $O && !AllowO ($OWNR{$h}, $O, $PROF{$h}) || $USER ne "root" && !AllowO $OWNR{$h}, $USER, $PROF{$h};
   @q = split /\s+/, $QHOST{$h};
   (!$du{$h} || !$dv{$h} || $_f) && ($du{$h} = GetKb ("$HOME{$PROF{$h}}/$h"), $dv{$h} = GetKb ("$HOME{$PROF{$h}}/$h/var"), 
      $DUCACHE && `echo "$du{$h} $dv{$h}\t$h" $NERR >> /tmp/.vhost/.duh`), 
   $hl[$hl++] = sprintf "%-13s %1s %4s %4s %2s/%02d %2s/%02d %3s-%02d-%02d " . ($_l? "%-31s %-15s %-s\n" : "%-s\n"), 
      ($H? $OWNR{$h} : substr $OWNR{$h}, 0, 13) || "----", GetSh ($OWNR{$h}), $du{$h}, $dv{$h}, 
      -c "/etc/vdata/vusers.$h"? "*" : scalar grep (/^\s*[^#:]+:/, Cat "/etc/vdata/passwd.$h"), $q[1], 
      -c "/etc/vdata/vusers.$h"? "*" : scalar grep (/^\s*[^#:]+:/, Cat "/etc/vdata/aliases.$h"), $q[2], 
      $date[1], $date[2], substr ($date[5], -2), $h . (!CheckQ ($QHOST{$h}) && "-"), $HOST{$h}, $PROF{$h} if $h eq $3;
   $al[$al++] = sprintf "%-47s %-s\n", $h, $3 . (!CheckQ ($QHOST{$h}) && "-") if $h ne $3;
}
eval "\@HLIST = $hsort \@hl;";
eval "\@ALIST = $asort \@al;";


};
$X eq "lsuser" && do { #######################################################

my ($u, $ul, $vl, $usort, $vsort, $unum, $vnum, $ucmp, $vcmp, $umya, $vmya, $umyb, $vmyb, $vali, $vcall, @u, @q, @ul, @vl, @qu);
@qu = $H && Quota::query (Quota::getqcarg ($VHOME), $O[2]);
$unum = (0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10)[$_n<11?$_n:11];
$vnum = (0, 0, 1, 2, 3, 4, 1, 6, 7, 8, 9, 10)[$_n<11?$_n:11];
$umya = "(split /[\\/\\s]+/, \$a, 10)[$unum]", $umyb = "(split /[\\/\\s]+/, \$b, 10)[$unum]";
$vmya = "(split /[\\/\\s]+/, \$a,  2)[$vnum]", $vmyb = "(split /[\\/\\s]+/, \$b,  2)[$vnum]";
$umya = "GetNb ($umya)", $umyb = "GetNb ($umyb)" if $_n =~ /^[34]$/;
$ucmp = $H && $_n =~ /^[34]$/ || !$H && $_n =~ /^[345678]$/? "<=>" : "cmp";
$vcmp = "cmp";
$usort = ($_r && "reverse ") . "sort {$umya $ucmp $umyb} " if "$_s$_r$_n";
$vsort = ($_r && "reverse ") . "sort {$vmya $vcmp $vmyb} " if "$_s$_r$_n" && $vnum < 2;
$vcall = <<'#' if $vnum != 1;
my $i;
/^(catch-all\@\s.*)/s && splice (@VLIST, $i, 1) && 
($_r && $vnum < 2 && ($VLIST[$#VLIST+1] = $1) || unshift (@VLIST, $1)) && last, $i++ for @VLIST;
#

for (!$H && @USER)
{
   /^\s*([^#\s]+)\s+(\S+)\s.*\d\s*(\svroot)?\s*$/ && (@u = getpwnam $1, $u = $1) || next;
   next if $O && !AllowO ($u, $O, $2) || $USER ne "root" && !AllowO $u, $USER, $2;
   @q = split /\s+/, $QUSER{$u};
   @qu = @u && Quota::query (Quota::getqcarg ($HOME{$2}), $u[2]);
   $ul[$ul++] = sprintf "%-13s %1s %4s/%4s %2s/%02d %2s/%02d %5s %-s\n", $O? $u : substr ($u, 0, 13), 
      GetSh ($u), GetKb ($qu[0] * $blockSize), GetKb ($qu[2] * $blockSize), 
      scalar grep ($OWNR{$_} eq $u, keys %PROF), $q[1], GetAc ($u), $q[2], $3 && "vroot" || "  -  ", $2;
}
for ($H && @VPWD)
{
   /^\s*([^#:]+):([^:]*:){3}([^:]*):/ && (@u = getpwnam $1, $u = $1) || next;
   next if $LOH && !@u || $LOH && $u[3] == $GUID && $u[6] eq "vhost vuser";
   $ul[$ul++] = sprintf "%-58s %-s\n", $LOH? 
      (sprintf ("%-13s %1s %4s %4s %-s", $u, GetSh ($u), $u[7] eq "$homeDir/$u"? GetKb $u[7] : "----", 
      GetKb ("$mailDir/$u"), "----"), $u[6]) : 
      (sprintf ("%-13s %1s %4s %4s %-s", $u, GetSh ($u, $u ne $O), GetKb ("$VHOME/$H/home/$u"), 
      GetKb ($u ne $O? "$VHOME/$H/var/mail/$u" : "$mailDir/$u"), $VUSR{$u} || "----"), $3);
}
for ($H && @VALI)
{
   /^\s*([^#:]+):/ && ($u = $1) || next;
   chomp ($vali = $VALI{$u});
   $vali =~ s/\s*,\s*/ /g;
   $vl[$vl++] = sprintf "%-13s %-s\n", $u, $vali;
}
eval "\@ULIST = $usort \@ul;";
eval "\@VLIST = $vsort \@vl; $vcall \@VLIST;" if $H;


};
$X eq "hostlog" && do { ######################################################

my $l;
my $log = (split /\s+/, `ls $VHOME/$H/var/log/ $NERR`)[0];
print STDERR `ls -l $VHOME/$H/var/log/ $NERR`, "log [$log]: " if !$_q;
print STDOUT `ls -l $VHOME/$H/var/log/ $NERR`, "log [$log]: " if $_q;
print STDERR $l = $L, "\n" if $L;
print STDOUT $l = $L, "\n" if $_q;
chomp ($l = <STDIN>) if !$L && !$_q;
$l = $l || $log;
Error5 "'$VHOME/$H/var/log/$l' is not a log file" if !-f "$VHOME/$H/var/log/$l" && !$_q;

system $l =~ /\.gz$/? "gunzip -c $VHOME/$H/var/log/$l $NERR | $PAGER" : "$PAGER $VHOME/$H/var/log/$l" if !$_q;


};
$X eq "cmdlog" && do { #######################################################

my $l;
my $g = $G || $CMDLOG;
Error5 "'$g' is not a log file" if !-f $g;

for ($g =~ /\.gz$/? `gunzip -c $g $NERR` : Cat $g)
{
   / vhost\[\d+:\d+\]: \(([^(\@)]+)\@?([^(\@)]*)\) --/ || next;
   next if !$HOST && $USER ne "root" && (!$2 && !AllowO ($1) || $2 && !AllowO $OWNR{$2}) || 
      $HOST && ($2 ne $HOST || $1 ne $USER);
   s/([\\\$"`])/\\$1/g;
   $l .= $_;
}
Flock "/tmp/.vhost/.out.$$", ">", 0;
print FILE $l;
Flock 0;
system "(cat /tmp/.vhost/.out.$$; rm /tmp/.vhost/.out.$$) $NERR | $PAGER";


};
$X eq "reload" && do { #######################################################

$LOCAL = 1, Rexec "vhost --reload ".($_n !~ /^\s*$/ && "-$_n ").$_f, "###", "$HOSTs $DNSs" if !$LOCAL && !$_l;

exit 0 if $_n && fork;
sleep $_n if $_n;
print "\n" if $_n;
ReloadRc $dnsRc;
ReloadRc 0 if $_f && $USER eq "root";


};
$X eq "rotatelog" && do { ####################################################

my $ext0 = ".gz" if !$_z;
my $ext1 = ".gz" if $_z;
my $gzip = $_z? "gzip" : "gunzip";
for my $h ($H || grep CheckQ ($QHOST{$_}), keys %PROF)
{
   my $d = "$HOME{$PROF{$h}}/$h/var/log";
   for my $f (grep Cat ("$d/.no-rotate") !~ /^\s*\Q$_\E\s*$/m && !/\.\d+(\.gz)?$/ && -f "$d/$_", split /\s+/, `ls $d/ $NERR`)
   {
      my ($i, $j);
      $i = -$_, `$gzip -f $d/$f.$i$ext0 $NERR` for -$_n..-1;
      $i = -$_, $j = -$_-1, `cp -p $d/$f.$j$ext1 $d/$f.$i$ext1 $NERR` for -$_n..-2;
      `cp -p $d/$f $d/$f.1 $NERR` if $_n;
      `$gzip -f $d/$f.1 $NERR` if $_n;
      Flock "$d/$f", ">", 0;
      Flock 0;
   }
}


};
$X eq "getvalue" && do { #####################################################

print "$_\n" for eval $Y;


};
$X eq "uninstall" && do { ####################################################

my $version = (split /\s+/, $VERSION)[1];
Prompt "uninstalling vhost $version";

`mv /usr/local/vhost/bin /tmp/.vhost $NERR`;
`rm -r /usr/local/vhost/* $NERR`;
`mv /tmp/.vhost/bin /usr/local/vhost $NERR`;
`rm -r /tmp/.vhost $NERR`;
Flock "$cronDir/root", "+<", 0 if grep /^\s*$FOOT/, Cat "$cronDir/root";
s/^\s*$FOOT.*//s, s/^[^#]*\/usr\/sbin\/vhost\s.*//s for @FILE;
print FILE @FILE;
`touch $cronDir $NERR` if -f FILE;
Flock 0;


};
$X eq "finalization" && do { #################################################

`rm /tmp/.vhost/.ins.autoconf $NERR` if -f "/tmp/.vhost/.ins.autoconf";

if ($ARGV[0] eq "addhost")
{
   `cp -pR $HTEMP/. $VHOME/$H $NERR` if -d $HTEMP;
   `cp -pR $UTEMP/. $VHOME/$H/home/$O $NERR` if -d $UTEMP;
   `chown -R $O[2]:$GHID $VHOME/$H $NERR`;
   `chown 0:$GHID $VHOME/$H/var $NERR`;
   `chown 0:$GHID $VHOME/$H/var/log $NERR`;
   `chown 0:$GHID $VHOME/$H/var/mail $NERR`;
   `chmod 0750 $VHOME/$H $NERR`;
   `touch $VHOME/$H $NERR`;
   `touch $VHOME/$H/home/$O $NERR`;
   -l "$VHOME/$H/var/mail/$_" && `chown 0:$GHID $VHOME/$H/var/mail/$_ $NERR`, 
   -l "$VHOME/$H/home/$_" && `chown 0:$GHID $VHOME/$H/home/$_ $NERR` 
      for split /\s+/, `ls $VHOME/$H/var/mail/ $NERR`;
}
if ($ARGV[0] eq "adduser")
{
   `cp -pR $UTEMP/. $VHOME/$H/home/$U $NERR` if -d $UTEMP;
   `chown -R $U[2]:$U[3] $homeDir/$U $NERR` if $LOH;
   `chown 0:$GHID $VHOME/$H/home/$U $NERR` if $LOH;
   `chown 0:$GHID $VHOME/$H/var/mail/$U $NERR` if $LOH;
   `chmod 0755 $homeDir/$U $NERR` if $LOH;
   `touch $homeDir/$U $NERR` if $LOH;
   `chown -R $O[2]:$GHID $VHOME/$H/home/$U $NERR` if !$LOH;
   `chown $O[2]:$GHID $VHOME/$H/var/mail/$U $NERR` if !$LOH;
   `chmod 0755 $VHOME/$H/home/$U $NERR` if !$LOH;
   `touch $VHOME/$H/home/$U $NERR` if !$LOH;
}
if ($ARGV[0] eq "lshost")
{
   printf "# OWNER       S USED VAR/ #U/QU #A/QA MON-DT-YR HOST - host total %d\n", $#HLIST+1;
   print @HLIST;
   printf "# - host alias total %d\n", $#ALIST+1;
   print @ALIST;
}
if ($ARGV[0] eq "lsuser")
{
   printf "# OWNER       S USED/QUOT #H/QH #A/QA VROOT PROFILE - user total %d\n", $#ULIST+1 if !$H;
   printf "# USER        S HOME MAIL ALTERNATIVE_LOGIN    FULL_NAME - user total %d\n", $#ULIST+1 if $H;
   print @ULIST;
   printf "# - user alias total %d\n", $#VLIST+1 if $H;
   print @VLIST if $H;
}
if ($ARGV[0] eq "uninstall")
{
   `rm /usr/sbin/vhost $NERR`;
   `rm /usr/sbin/vhost.pl $NERR`;
}
ReloadRc $inetdRc if $ARGV[0] eq "reload";


};
}
